# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import errno
import os
import resource
import unittest

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn import cast
from drgn.helpers.linux.bpf import (
    bpf_btf_for_each,
    bpf_link_for_each,
    bpf_map_by_id,
    bpf_map_for_each,
    bpf_prog_by_id,
    bpf_prog_for_each,
    bpf_prog_used_maps,
    cgroup_bpf_prog_for_each,
    cgroup_bpf_prog_for_each_effective,
)
from drgn.helpers.linux.cgroup import cgroup_get_from_path
from drgn.helpers.linux.fs import fget
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import LinuxKernelTestCase
from tests.linux_kernel.bpf import (
    BPF_CGROUP_INET_INGRESS,
    BPF_EXIT_INSN,
    BPF_F_ALLOW_MULTI,
    BPF_LD_MAP_FD,
    BPF_MAP_TYPE_HASH,
    BPF_MOV64_IMM,
    BPF_PROG_TYPE_CGROUP_SKB,
    BPF_PROG_TYPE_SOCKET_FILTER,
    BPF_REG_0,
    _SYS_bpf,
    bpf_btf_ids,
    bpf_link_create,
    bpf_link_ids,
    bpf_map_create,
    bpf_map_get_info_by_fd,
    bpf_map_ids,
    bpf_prog_attach,
    bpf_prog_get_info_by_fd,
    bpf_prog_ids,
    bpf_prog_load,
)
from tests.linux_kernel.helpers.test_cgroup import tmp_cgroups


class BpfTestCase(LinuxKernelTestCase):
    INSNS = (
        BPF_MOV64_IMM(BPF_REG_0, 0),
        BPF_EXIT_INSN(),
    )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if _SYS_bpf is None:
            raise unittest.SkipTest(
                f"bpf syscall number is not known on {NORMALIZED_MACHINE_NAME}"
            )
        # Before the patch series culminating in Linux kernel commit
        # 3ac1f01b43b6 ("bpf: Eliminate rlimit-based memory accounting for bpf
        # progs") (in v5.11), BPF program and map memory usage was limited by
        # RLIMIT_MEMLOCK. At that time (before Linux kernel commit 9dcc38e2813e
        # ("Increase default MLOCK_LIMIT to 8 MiB") (in v5.16)), the limit was
        # only 64kB. We only allocate a few small objects at a time, but with
        # 64k pages, we can easily blow that limit.
        memlock_limit = 8 * 1024 * 1024
        old_limit = resource.getrlimit(resource.RLIMIT_MEMLOCK)
        if old_limit[0] < memlock_limit:
            resource.setrlimit(
                resource.RLIMIT_MEMLOCK,
                (memlock_limit, max(memlock_limit, old_limit[1])),
            )
            cls.addClassCleanup(resource.setrlimit, resource.RLIMIT_MEMLOCK, old_limit)
        try:
            os.close(bpf_map_create(BPF_MAP_TYPE_HASH, 8, 8, 8))
        except OSError as e:
            if e.errno != errno.ENOSYS:
                raise
            raise unittest.SkipTest(
                "kernel does not support bpf syscall (CONFIG_BPF_SYSCALL)"
            )


class TestBpf(BpfTestCase):
    def test_bpf_btf_for_each(self):
        # BTF was added in Linux kernel commit 69b693f0aefa ("bpf: btf:
        # Introduce BPF Type Format (BTF)") (in v4.18) and had IDs from the
        # start, but there was no API to get them until commit 1b9ed84ecf26
        # ("bpf: add new BPF_BTF_GET_NEXT_ID syscall command") (in v5.4). The
        # only kernel version that we support in between is v4.19, which we can
        # live without testing.
        #
        # Note that before Linux kernel commits 5329722057d4 ("bpf: Assign ID
        # to vmlinux BTF and return extra info for BTF in GET_OBJ_INFO") and
        # 36e68442d1af ("bpf: Load and verify kernel module BTFs") (in v5.11),
        # there won't be any BTF IDs unless they were explicitly added outside
        # of the test suite.
        try:
            expected_ids = list(bpf_btf_ids())
        except OSError as e:
            if e.errno != errno.EINVAL:
                raise
            self.skipTest("kernel does not support BPF_BTF_GET_NEXT_ID")
        self.assertCountEqual(
            [btf.id.value_() for btf in bpf_btf_for_each(self.prog)], expected_ids
        )

    def test_bpf_link_for_each(self):
        with tmp_cgroups() as cgroups:
            fds = []
            try:
                for cgroup in cgroups:
                    fds.append(os.open(cgroup, os.O_RDONLY | os.O_DIRECTORY))
                for i in range(3):
                    # Cgroup BPF programs didn't exist before Linux kernel
                    # commit 3007098494be ("cgroup: add support for eBPF
                    # programs") (in v4.10).
                    try:
                        prog_fd = bpf_prog_load(
                            BPF_PROG_TYPE_CGROUP_SKB,
                            self.INSNS,
                            b"GPL",
                            expected_attach_type=BPF_CGROUP_INET_INGRESS,
                        )
                        fds.append(prog_fd)
                    except OSError as e:
                        if e.errno != errno.EINVAL:
                            raise
                        self.skipTest(
                            "kernel does not support BPF_PROG_TYPE_CGROUP_SKB"
                        )
                    # BPF links didn't exist before Linux kernel commit
                    # 70ed506c3bbc ("bpf: Introduce pinnable bpf_link
                    # abstraction") (in v5.7).
                    try:
                        fds.append(
                            bpf_link_create(
                                prog_fd, fds[i % len(cgroups)], BPF_CGROUP_INET_INGRESS
                            )
                        )
                    except OSError as e:
                        if e.errno != errno.EINVAL:
                            raise
                        self.skipTest("kernel does not support BPF_LINK_CREATE")

                # bpf_link_for_each() isn't supported before Linux v5.8, which
                # added IDs for BPF links in commit a3b80e107894 ("bpf:
                # Allocate ID for bpf_link") and an API to get them in commit
                # 2d602c8cf40d ("bpf: Support GET_FD_BY_ID and GET_NEXT_ID for
                # bpf_link").
                try:
                    expected_ids = list(bpf_link_ids())
                except OSError as e:
                    if e.errno != errno.EINVAL:
                        raise
                    self.skipTest("kernel does not support BPF_LINK_GET_NEXT_ID")
                self.assertCountEqual(
                    [link.id.value_() for link in bpf_link_for_each(self.prog)],
                    expected_ids,
                )
            finally:
                for fd in fds:
                    os.close(fd)

    def test_bpf_map_for_each(self):
        fds = []
        try:
            for i in range(3):
                fds.append(bpf_map_create(BPF_MAP_TYPE_HASH, 8, 8, 8))

            # bpf_map_for_each() isn't supported before Linux v4.13, which
            # added IDs for BPF maps in commit f3f1c054c288 ("bpf: Introduce
            # bpf_map ID") and a API to get them in commit 34ad5580f8f9 ("bpf:
            # Add BPF_(PROG|MAP)_GET_NEXT_ID command").
            try:
                expected_ids = list(bpf_map_ids())
            except OSError as e:
                if e.errno != errno.EINVAL:
                    raise
                self.skipTest("kernel does not support BPF_MAP_GET_NEXT_ID")

            self.assertCountEqual(
                [map.id.value_() for map in bpf_map_for_each(self.prog)], expected_ids
            )
        finally:
            for fd in fds:
                os.close(fd)

    def test_bpf_prog_for_each(self):
        fds = []
        try:
            for i in range(3):
                fds.append(
                    bpf_prog_load(BPF_PROG_TYPE_SOCKET_FILTER, self.INSNS, b"GPL")
                )

            # bpf_prog_for_each() isn't supported before Linux v4.13, which
            # added IDs for BPF programs in commit dc4bb0e23561 ("bpf:
            # Introduce bpf_prog ID") and an API to get them in commit
            # 34ad5580f8f9 ("bpf: Add BPF_(PROG|MAP)_GET_NEXT_ID command").
            try:
                expected_ids = list(bpf_prog_ids())
            except OSError as e:
                if e.errno != errno.EINVAL:
                    raise
                self.skipTest("kernel does not support BPF_PROG_GET_NEXT_ID")

            self.assertCountEqual(
                [prog.aux.id.value_() for prog in bpf_prog_for_each(self.prog)],
                expected_ids,
            )
        finally:
            for fd in fds:
                os.close(fd)

    def test_bpf_prog_used_maps(self):
        with contextlib.ExitStack() as exit_stack:
            map_fd = bpf_map_create(BPF_MAP_TYPE_HASH, 8, 8, 8)
            exit_stack.callback(os.close, map_fd)

            prog_fd = bpf_prog_load(
                BPF_PROG_TYPE_SOCKET_FILTER,
                BPF_LD_MAP_FD(BPF_REG_0, map_fd) + self.INSNS,
                b"GPL",
            )
            exit_stack.callback(os.close, prog_fd)

            bpf_prog = cast(
                "struct bpf_prog *",
                fget(find_task(self.prog, os.getpid()), prog_fd).private_data,
            )

            bpf_map = cast(
                "struct bpf_map *",
                fget(find_task(self.prog, os.getpid()), map_fd).private_data,
            )

            self.assertEqual(list(bpf_prog_used_maps(bpf_prog)), [bpf_map])

    def test_cgroup_bpf_prog_for_each(self):
        with tmp_cgroups() as (parent_cgroup, child_cgroup):
            fds = []
            try:
                parent_cgroup_fd = os.open(parent_cgroup, os.O_RDONLY | os.O_DIRECTORY)
                fds.append(parent_cgroup_fd)
                child_cgroup_fd = os.open(child_cgroup, os.O_RDONLY | os.O_DIRECTORY)
                fds.append(child_cgroup_fd)
                try:
                    parent_prog_fd = bpf_prog_load(
                        BPF_PROG_TYPE_CGROUP_SKB,
                        self.INSNS,
                        b"GPL",
                        expected_attach_type=BPF_CGROUP_INET_INGRESS,
                    )
                    fds.append(parent_prog_fd)
                except OSError as e:
                    if e.errno != errno.EINVAL:
                        raise
                    # If the kernel doesn't support cgroup BPF programs, the
                    # helpers should return empty lists.
                    parent_ids = child_ids = child_effective_ids = []
                else:
                    parent_prog_id = bpf_prog_get_info_by_fd(parent_prog_fd).id
                    parent_ids = [parent_prog_id]

                    # If the kernel supports BPF_F_ALLOW_MULTI, test with
                    # multiple programs.
                    try:
                        bpf_prog_attach(
                            parent_cgroup_fd,
                            parent_prog_fd,
                            BPF_CGROUP_INET_INGRESS,
                            attach_flags=BPF_F_ALLOW_MULTI,
                        )
                    except OSError as e:
                        if e.errno != errno.EINVAL:
                            raise
                        bpf_prog_attach(
                            parent_cgroup_fd, parent_prog_fd, BPF_CGROUP_INET_INGRESS
                        )
                        child_ids = []
                        child_effective_ids = [parent_prog_id]
                    else:
                        child_prog_fd = bpf_prog_load(
                            BPF_PROG_TYPE_CGROUP_SKB,
                            self.INSNS,
                            b"GPL",
                            expected_attach_type=BPF_CGROUP_INET_INGRESS,
                        )
                        fds.append(child_prog_fd)
                        child_prog_id = bpf_prog_get_info_by_fd(child_prog_fd).id
                        bpf_prog_attach(
                            child_cgroup_fd, child_prog_fd, BPF_CGROUP_INET_INGRESS
                        )
                        child_ids = [child_prog_id]
                        child_effective_ids = [parent_prog_id, child_prog_id]

                parent_cgrp = cgroup_get_from_path(self.prog, parent_cgroup.name)
                child_cgrp = cgroup_get_from_path(
                    self.prog, parent_cgroup.name + "/" + child_cgroup.name
                )

                self.assertCountEqual(
                    [
                        prog.aux.id.value_()
                        for prog in cgroup_bpf_prog_for_each(
                            parent_cgrp, BPF_CGROUP_INET_INGRESS
                        )
                    ],
                    parent_ids,
                )
                self.assertCountEqual(
                    [
                        prog.aux.id.value_()
                        for prog in cgroup_bpf_prog_for_each(
                            child_cgrp, BPF_CGROUP_INET_INGRESS
                        )
                    ],
                    child_ids,
                )

                self.assertCountEqual(
                    [
                        prog.aux.id.value_()
                        for prog in cgroup_bpf_prog_for_each_effective(
                            parent_cgrp, BPF_CGROUP_INET_INGRESS
                        )
                    ],
                    parent_ids,
                )
                self.assertCountEqual(
                    [
                        prog.aux.id.value_()
                        for prog in cgroup_bpf_prog_for_each_effective(
                            child_cgrp, BPF_CGROUP_INET_INGRESS
                        )
                    ],
                    child_effective_ids,
                )
            finally:
                for fd in fds:
                    os.close(fd)

    def test_bpf_prog_by_id(self):
        fd = bpf_prog_load(BPF_PROG_TYPE_SOCKET_FILTER, self.INSNS, b"GPL")
        self.addCleanup(os.close, fd)
        try:
            prog_id = bpf_prog_get_info_by_fd(fd).id
        except OSError as e:
            # bpf_prog_by_id() and BPF program IDs aren't supported before
            # Linux v4.13, which added IDs for BPF programs in commit
            # dc4bb0e23561 ("bpf: Introduce bpf_prog ID") and an API to get
            # them in commit 34ad5580f8f9 ("bpf: Add BPF_(PROG|MAP)_GET_NEXT_ID
            # command").
            if e.errno != errno.EINVAL:
                raise
            self.skipTest("kernel does not support BPF_PROG_GET_NEXT_ID")

        prog = bpf_prog_by_id(self.prog, prog_id)
        self.assertNotEqual(prog.value_(), 0)
        self.assertEqual(prog.aux.id.value_(), prog_id)

        non_existent_prog = bpf_prog_by_id(self.prog, 0x7FFFFFFF)
        self.assertEqual(non_existent_prog.value_(), 0)

    def test_bpf_map_by_id(self):
        fd = bpf_map_create(BPF_MAP_TYPE_HASH, 8, 8, 8)
        self.addCleanup(os.close, fd)
        try:
            map_id = bpf_map_get_info_by_fd(fd).id
        except OSError as e:
            # bpf_map_by_id() and BPF map IDs aren't supported before
            # Linux v4.13, which added IDs for BPF maps in commit
            # f3f1c054c288 ("bpf: Introduce bpf_map ID") and an API to get
            # them in commit 34ad5580f8f9 ("bpf: Add BPF_(PROG|MAP)_GET_NEXT_ID
            # command").
            if e.errno != errno.EINVAL:
                raise
            self.skipTest("kernel does not support BPF_MAP_GET_NEXT_ID")

        bpf_map = bpf_map_by_id(self.prog, map_id)
        self.assertNotEqual(bpf_map.value_(), 0)
        self.assertEqual(bpf_map.id.value_(), map_id)

        non_existent_map = bpf_map_by_id(self.prog, 0x7FFFFFFF)
        self.assertEqual(non_existent_map.value_(), 0)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import logging
import os
import os.path
from pathlib import Path
import re
import shlex
import shutil
import subprocess
import sys
import sysconfig

from setuptools import Command, find_packages, setup
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.egg_info import egg_info as _egg_info
from setuptools.command.sdist import sdist as _sdist
from setuptools.extension import Extension

# setuptools must be imported before distutils (see pypa/setuptools#2230), so
# make sure to keep these fallbacks after the other setuptools imports.
try:
    # This was added in setuptools 62.4.0 (released June 13th, 2022).
    from setuptools.command.build import build as _build
except ImportError:
    from distutils.command.build import build as _build
try:
    # This was added in setuptools 59.0.0 (released November 12th, 2021).
    from setuptools.errors import BaseError
except ImportError:
    from distutils.errors import DistutilsError as BaseError

from util import nproc
from vmtest.config import KERNEL_FLAVORS, SUPPORTED_KERNEL_VERSIONS, _run_autoreconf

logger = logging.getLogger(__name__)


class build(_build):
    def finalize_options(self):
        if self.build_temp is None or self.build_platlib is None:
            # Set build_temp/build_pathlib based on the extension suffix rather
            # than the default (which is based on sysconfig.get_platform() and
            # sys.implementation.cache_tag) to work around old setuptools
            # versions missing commit 01e5f2d36476 ("Use a separate build
            # directory for free-threading"), and also to make it easier to
            # coordinate with cross-compilation that doesn't use setuptools.
            suffix = sysconfig.get_config_var("EXT_SUFFIX")
            if suffix.endswith(".so"):
                suffix = suffix[:-3]
            if self.build_temp is None:
                self.build_temp = os.path.join(self.build_base, "temp" + suffix)
            if self.build_platlib is None:
                self.build_platlib = os.path.join(self.build_base, "lib" + suffix)

        super().finalize_options()

        if self.parallel is None:
            self.parallel = nproc() + 1


class build_ext(_build_ext):
    user_options = [
        ("inplace", "i", "put compiled extension into the source directory"),
        ("parallel=", "j", "number of parallel build jobs"),
    ]

    boolean_options = ["inplace"]

    help_options = []

    def _run_configure(self):
        self.mkpath(self.build_temp)
        makefile = os.path.join(self.build_temp, "Makefile")
        if not os.path.exists(makefile):
            args = [
                os.path.relpath("libdrgn/configure", self.build_temp),
                "--disable-static",
                "--disable-libdrgn",
                "--enable-python-extension",
            ]
            try:
                args.extend(shlex.split(os.environ["CONFIGURE_FLAGS"]))
            except KeyError:
                pass
            try:
                subprocess.check_call(
                    args,
                    cwd=self.build_temp,
                    env={**os.environ, "PYTHON": sys.executable},
                )
            except Exception:
                with contextlib.suppress(FileNotFoundError):
                    os.remove(makefile)
                raise

    def _run_make(self, *make_args):
        args = ["make", "-C", self.build_temp]
        if self.parallel:
            args.append(f"-j{self.parallel}")
        args.extend(make_args)
        subprocess.check_call(args)

    def make(self, *make_args):
        _run_autoreconf()
        self._run_configure()
        self._run_make(*make_args)

    def run(self):
        self.make()
        so = os.path.join(self.build_temp, ".libs/_drgn.so")
        if self.inplace:
            self.copy_file(so, self.get_ext_fullpath("_drgn"))
        old_inplace, self.inplace = self.inplace, 0
        build_path = self.get_ext_fullpath("_drgn")
        self.mkpath(os.path.dirname(build_path))
        self.copy_file(so, build_path)
        self.inplace = old_inplace


# Work around pypa/setuptools#436.
class egg_info(_egg_info):
    def run(self):
        if os.path.exists(".git"):
            with contextlib.suppress(FileNotFoundError):
                os.remove(os.path.join(self.egg_info, "SOURCES.txt"))
        super().run()


class sdist(_sdist):
    def make_release_tree(self, base_dir, files):
        super().make_release_tree(base_dir, files)
        # Add the libdrgn distribution tree. This won't add the file names to
        # .egg-info/SOURCES.txt, but as far as I can tell that doesn't matter.
        build_ext = self.get_finalized_command("build_ext")
        distdir = os.path.join(
            os.path.relpath(base_dir, build_ext.build_temp), "libdrgn"
        )
        build_ext.make("distdir", "distdir=" + distdir)


class test(Command):
    description = "run unit tests after in-place build"

    user_options = [
        (
            "kernel",
            "K",
            "run Linux kernel tests in a virtual machine on all supported kernels "
            f"({', '.join(SUPPORTED_KERNEL_VERSIONS)})",
        ),
        (
            "flavor=",
            "f",
            "when combined with -K, run Linux kernel tests on a specific flavor "
            f"({', '.join(KERNEL_FLAVORS)}) instead of the default flavor",
        ),
        (
            "all-kernel-flavors",
            "F",
            "when combined with -K, run Linux kernel tests on all supported flavors "
            f"({', '.join(KERNEL_FLAVORS)}) instead of just the default flavor",
        ),
        (
            "extra-kernels=",
            "k",
            "additional kernels to run Linux kernel helper tests on in a virtual machine "
            "(comma-separated list of wildcard patterns matching uploaded kernel releases)",
        ),
        (
            "vmtest-dir=",
            "d",
            "directory for built artifacts and downloaded kernels for virtual machine tests (default: 'build/vmtest')",
        ),
    ]

    def initialize_options(self):
        self.kernel = False
        self.flavor = "default"
        self.all_kernel_flavors = False
        self.extra_kernels = ""
        self.vmtest_dir = None

    def finalize_options(self):
        self.kernels = [kernel for kernel in self.extra_kernels.split(",") if kernel]
        if self.kernel:
            flavors = KERNEL_FLAVORS if self.all_kernel_flavors else [self.flavor]
            self.kernels.extend(
                kernel + ".*" + flavor
                for kernel in SUPPORTED_KERNEL_VERSIONS
                for flavor in flavors
            )
        if self.vmtest_dir is None:
            build_base = self.get_finalized_command("build").build_base
            self.vmtest_dir = os.path.join(build_base, "vmtest")

    def _run_local(self):
        import unittest

        try:
            self.get_finalized_command("build_ext").make("check")
            make_check_success = True
        except subprocess.CalledProcessError:
            make_check_success = False

        argv = ["discover"]
        if self.verbose:
            argv.append("-v")
        test = unittest.main(module=None, argv=argv, exit=False)
        return make_check_success and test.result.wasSuccessful()

    def _run_vm(self, kernel):
        import vmtest.vm

        logger.info("running tests in VM on Linux %s", kernel.release)

        command = rf"""
set -e

export PYTHON={shlex.quote(sys.executable)}
if [ -e /proc/vmcore ]; then
    "$PYTHON" -Bm unittest discover -t . -s tests/linux_kernel/vmcore {"-v" if self.verbose else ""}
else
    insmod "$DRGN_TEST_KMOD"
    DRGN_RUN_LINUX_KERNEL_TESTS=1 "$PYTHON" -Bm \
        unittest discover -t . -s tests/linux_kernel {"-v" if self.verbose else ""}
    "$PYTHON" -Bm vmtest.enter_kdump
    # We should crash and not reach this.
    exit 1
fi
"""
        try:
            returncode = vmtest.vm.run_in_vm(
                command,
                kernel,
                Path("/"),
                Path(self.vmtest_dir),
                test_kmod=vmtest.vm.TestKmodMode.BUILD,
            )
        except vmtest.vm.LostVMError:
            logger.exception("error on Linux %s", kernel.release)
            return False
        logger.info("Tests in VM on Linux %s returned %d", kernel.release, returncode)
        return returncode == 0

    def run(self):
        import urllib.error

        from vmtest.config import ARCHITECTURES, Kernel
        from vmtest.download import DownloadCompiler, DownloadKernel, download_in_thread

        in_github_actions = os.getenv("GITHUB_ACTIONS") == "true"

        if in_github_actions:

            @contextlib.contextmanager
            def github_workflow_group(title):
                sys.stdout.flush()
                print("::group::" + title, file=sys.stderr, flush=True)
                try:
                    yield
                finally:
                    sys.stdout.flush()
                    print("::endgroup::", file=sys.stderr, flush=True)

        else:

            @contextlib.contextmanager
            def github_workflow_group(title):
                yield

        # Start downloads ASAP so that they're hopefully done by the time we
        # need them.
        try:
            to_download = []
            if self.kernels:
                to_download.append(DownloadCompiler(ARCHITECTURES["x86_64"]))
                for pattern in self.kernels:
                    to_download.append(DownloadKernel(ARCHITECTURES["x86_64"], pattern))

            # Downloading too many files before they can be used for testing runs the
            # risk of filling up the limited disk space is Github Actions. Set a limit
            # of no more than 5 files which can be downloaded ahead of time. This is a
            # magic number which is inexact, but works well enough.
            max_pending_kernels = 5 if in_github_actions else 0

            with download_in_thread(
                Path(self.vmtest_dir), to_download, max_pending_kernels
            ) as downloads:
                downloads_it = iter(downloads)

                if to_download:
                    logger.info("downloading kernels in the background")

                with github_workflow_group("Build extension"):
                    self.run_command("egg_info")
                    self.reinitialize_command("build_ext", inplace=1)
                    self.run_command("build_ext")

                passed = []
                failed = []

                with github_workflow_group("Run unit tests"):
                    if self.kernels:
                        logger.info("running tests locally")
                    if self._run_local():
                        passed.append("local")
                    else:
                        failed.append("local")

                for pattern in self.kernels:
                    while True:
                        kernel = next(downloads_it)
                        if isinstance(kernel, Kernel):
                            break
                    with github_workflow_group(
                        f"Run integration tests on Linux {kernel.release}"
                    ):
                        if self._run_vm(kernel):
                            passed.append(kernel.release)
                        else:
                            failed.append(kernel.release)

                    if passed:
                        logger.info("Passed: %s", ", ".join(passed))
                    if failed:
                        logger.error("Failed: %s", ", ".join(failed))

                    # Github Actions has limited disk space. Once tested, we
                    # will not use the kernel again, so delete it.
                    if in_github_actions:
                        logger.info("Deleting kernel %s", kernel.release)
                        shutil.rmtree(kernel.path)
        except urllib.error.HTTPError as e:
            if e.code == 403:
                print(e, file=sys.stderr)
                print("Headers:", e.headers, file=sys.stderr)
                print("Body:", e.read().decode(), file=sys.stderr)
            raise

        if failed:
            raise BaseError("some tests failed")
        else:
            logger.info("all tests passed")


def get_version():
    try:
        with open("drgn/internal/version.py", "r") as f:
            version_py = f.read()
    except FileNotFoundError:
        version_py = None

    # The public version always comes from configure.ac.
    with open("libdrgn/configure.ac", "r") as f:
        public_version = re.search(
            r"AC_INIT\(\[libdrgn\], \[([^]]*)\]", f.read()
        ).group(1)
    # Default local version if we fail.
    local_version = "+unknown"

    # If this is a git repository, use a git-describe(1)-esque local version.
    # Otherwise, get the local version saved in the sdist.
    if os.path.exists(".git") and (
        subprocess.call(
            ["git", "--git-dir=.git", "rev-parse"], stderr=subprocess.DEVNULL
        )
        == 0
    ):
        # Read the Docs modifies the working tree (namely, docs/conf.py). We
        # don't want the documentation to display a dirty version, so ignore
        # modifications for RTD builds.
        dirty = os.getenv("READTHEDOCS") != "True" and bool(
            subprocess.check_output(
                ["git", "status", "-uno", "--porcelain"],
                # Use the environment variable instead of --no-optional-locks
                # to support Git < 2.14.
                env={**os.environ, "GIT_OPTIONAL_LOCKS": "0"},
            )
        )

        try:
            count = int(
                subprocess.check_output(
                    ["git", "rev-list", "--count", f"v{public_version}.."],
                    stderr=subprocess.DEVNULL,
                    universal_newlines=True,
                )
            )
        except subprocess.CalledProcessError:
            logger.warning("warning: v%s tag not found", public_version)
        else:
            if count == 0:
                local_version = "+dirty" if dirty else ""
            else:
                commit = subprocess.check_output(
                    ["git", "rev-parse", "--short", "HEAD"], universal_newlines=True
                ).strip()
                local_version = f"+{count}.g{commit}"
                if dirty:
                    local_version += ".dirty"
    else:
        if version_py is None:
            # This isn't a proper sdist (maybe a git archive).
            logger.warning("warning: drgn/internal/version.py not found")
        else:
            # The saved version must start with the public version.
            match = re.search(
                rf'^__version__ = "{re.escape(public_version)}([^"]*)"$',
                version_py,
                re.M,
            )
            if match:
                local_version = match.group(1)
            else:
                logger.warning("warning: drgn/internal/version.py is invalid")

    version = public_version + local_version
    # Update version.py if necessary.
    new_version_py = f'__version__ = "{version}"\n'
    if new_version_py != version_py:
        with open("drgn/internal/version.py", "w") as f:
            f.write(new_version_py)
    return version


with open("README.rst", "r") as f:
    long_description = f.read()


setup(
    name="drgn",
    version=get_version(),
    packages=find_packages(include=["drgn", "drgn.*", "_drgn_util", "_drgn_util.*"]),
    package_data={"drgn": ["../_drgn.pyi", "py.typed"]},
    # This is here so that setuptools knows that we have an extension; it's
    # actually built using autotools/make.
    ext_modules=[Extension(name="_drgn", sources=[])],
    cmdclass={
        "build": build,
        "build_ext": build_ext,
        "egg_info": egg_info,
        "sdist": sdist,
        "test": test,
    },
    entry_points={
        "console_scripts": [
            "drgn=drgn.cli:_main",
            "drgn-crash=drgn.internal.crashcli:_main",
        ],
        "drgn.plugins": [
            "builtin_commands=drgn.commands._builtin",
            "crash_commands=drgn.commands._crash",
        ],
    },
    python_requires=">=3.8",
    author="Omar Sandoval",
    author_email="osandov@osandov.com",
    description="Programmable debugger",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/osandov/drgn",
    project_urls={
        "Bug Tracker": "https://github.com/osandov/drgn/issues",
        "Documentation": "https://drgn.readthedocs.io",
    },
    license="LGPL-2.1-or-later",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Debuggers",
        "Topic :: System :: Operating System Kernels :: Linux",
    ],
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import importlib
import pkgutil
import pydoc
import types

import drgn
import drgn.helpers.common
import drgn.helpers.linux
from tests import TestCase


class TestDocs(TestCase):
    def test_render(self):
        pydoc.render_doc(drgn)

    def test_helper_exports(self):
        for package in (drgn.helpers.common, drgn.helpers.linux):
            for module_info in pkgutil.iter_modules(
                package.__path__, prefix=package.__name__ + "."
            ):
                with self.subTest(module=module_info.name):
                    submodule = importlib.import_module(module_info.name)
                    expected = set()
                    for name in dir(submodule):
                        attr = getattr(submodule, name)
                        # Documented functions defined in the given module
                        # should be in __all__. This won't catch non-function
                        # callables or re-exports, but it's close enough.
                        if (
                            isinstance(
                                attr, (types.FunctionType, types.BuiltinFunctionType)
                            )
                            and getattr(attr, "__module__", None) == module_info.name
                            and getattr(attr, "__doc__", None)
                        ):
                            expected.add(name)
                    missing = expected - set(getattr(submodule, "__all__", ()))
                    if missing:
                        self.fail(
                            f"{module_info.name}.__all__ is missing {sorted(missing)}"
                        )



# Copyright (c) ByteDance, Inc. and its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os

from drgn.helpers.linux.fs import path_lookup
from drgn.helpers.linux.net import get_net_ns_by_inode, netdev_get_by_name
from drgn.helpers.linux.tc import qdisc_lookup
from tests.linux_kernel import (
    LinuxKernelTestCase,
    skip_unless_have_pyroute2,
    temp_netns,
)


@skip_unless_have_pyroute2
class TestTc(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.name, cls.ns = cls.enterClassContext(temp_netns())

    def test_qdisc_lookup(self):
        import pyroute2

        try:
            self.ns.link("add", ifname="dummy0", kind="dummy")
        except pyroute2.NetlinkError:
            self.skipTest("kernel does not support dummy interface (CONFIG_DUMMY)")

        dummy = self.ns.link_lookup(ifname="dummy0")[0]

        # tc qdisc add dev dummy0 root handle 1: prio
        try:
            self.ns.tc(
                "add",
                kind="prio",
                index=dummy,
                handle="1:",
                # default TCA_OPTIONS for sch_prio, see [iproute2] tc/q_prio.c:prio_parse_opt()
                bands=3,
                priomap=[1, 2, 2, 2, 1, 2, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1],
            )
        except pyroute2.NetlinkError:
            self.ns.link("delete", ifname="dummy0")
            self.skipTest(
                "kernel does not support Multi Band Priority Queueing (CONFIG_NET_SCH_PRIO)"
            )
        # tc qdisc add dev dummy0 parent 1:1 handle 10: sfq
        try:
            self.ns.tc("add", kind="sfq", index=dummy, parent="1:1", handle="10:")
        except pyroute2.NetlinkError:
            self.ns.link("delete", ifname="dummy0")
            self.skipTest(
                "kernel does not support Stochastic Fairness Queueing (CONFIG_NET_SCH_SFQ)"
            )
        # tc qdisc add dev dummy0 parent 1:2 handle 20: tbf rate 20kbit buffer 1600 limit 3000
        try:
            self.ns.tc(
                "add",
                kind="tbf",
                index=dummy,
                parent="1:2",
                handle="20:",
                rate=2500,
                burst=1600,
                limit=3000,
            )
        except pyroute2.NetlinkError:
            self.ns.link("delete", ifname="dummy0")
            self.skipTest(
                "kernel does not support Token Bucket Filter (CONFIG_NET_SCH_TBF)"
            )
        # tc qdisc add dev dummy0 parent 1:3 handle 30: sfq
        self.ns.tc("add", kind="sfq", index=dummy, parent="1:3", handle="30:")
        # tc qdisc add dev dummy0 ingress
        try:
            self.ns.tc("add", kind="ingress", index=dummy)
        except pyroute2.NetlinkError:
            self.ns.link("delete", ifname="dummy0")
            self.skipTest(
                "kernel does not support ingress Qdisc (CONFIG_NET_SCH_INGRESS)"
            )

        inode = path_lookup(
            self.prog, os.path.realpath(f"/var/run/netns/{self.name}")
        ).dentry.d_inode
        netdev = netdev_get_by_name(get_net_ns_by_inode(inode), "dummy0")

        self.assertEqual(qdisc_lookup(netdev, 0x1).ops.id.string_(), b"prio")
        self.assertEqual(qdisc_lookup(netdev, 0x10).ops.id.string_(), b"sfq")
        self.assertEqual(qdisc_lookup(netdev, 0x20).ops.id.string_(), b"tbf")
        self.assertEqual(qdisc_lookup(netdev, 0x30).ops.id.string_(), b"sfq")
        self.assertEqual(qdisc_lookup(netdev, 0xFFFF).ops.id.string_(), b"ingress")

        self.ns.link("delete", ifname="dummy0")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import logging
import os
from pathlib import Path
import shutil
import subprocess
import tempfile
from typing import Optional, TextIO

from util import nproc, out_of_date
from vmtest.config import Kernel
from vmtest.download import downloaded_compiler

logger = logging.getLogger(__name__)


def build_kmod(
    download_dir: Path, kernel: Kernel, outfile: Optional[TextIO] = None
) -> Path:
    kmod = kernel.path.parent / f"drgn_test-{kernel.release}.ko"
    # External modules can't do out-of-tree builds for some reason, so copy the
    # source files to a temporary directory and build the module there, then
    # move it to the final location.
    kmod_source_dir = Path("tests/linux_kernel/kmod")
    source_files = ("drgn_test.c", "Makefile")
    if out_of_date(kmod, *[kmod_source_dir / filename for filename in source_files]):
        logger.info("building %s", kmod)

        compiler = downloaded_compiler(download_dir, kernel.arch)
        kernel_build_dir = kernel.path / "build"

        with tempfile.TemporaryDirectory(dir=kmod.parent) as tmp_name:
            tmp_dir = Path(tmp_name)
            # Make sure that header files have the same paths as in the
            # original kernel build.
            debug_prefix_map = [
                f"{kernel_build_dir.resolve()}=.",
                f"{tmp_dir.resolve()}=./drgn_test",
            ]
            cflags = " ".join(["-fdebug-prefix-map=" + map for map in debug_prefix_map])
            for filename in source_files:
                shutil.copy(kmod_source_dir / filename, tmp_dir / filename)
            subprocess.check_call(
                [
                    "make",
                    "ARCH=" + kernel.arch.kernel_arch,
                    "-C",
                    kernel_build_dir,
                    f"M={tmp_dir.resolve()}",
                    "KAFLAGS=" + cflags,
                    "KCFLAGS=" + cflags,
                    "-j",
                    str(nproc()),
                ],
                env={**os.environ, **compiler.env()},
                stdout=outfile,
                stderr=outfile,
            )
            (tmp_dir / "drgn_test.ko").rename(kmod)
    else:
        logger.info("%s is up to date", kmod)
    return kmod


def _main() -> None:
    import argparse

    from vmtest.download import (
        DOWNLOAD_KERNEL_ARGPARSE_METAVAR,
        DownloadCompiler,
        download_in_thread,
        download_kernel_argparse_type,
    )

    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", level=logging.INFO
    )

    parser = argparse.ArgumentParser(
        description="build drgn test kernel module",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-k",
        "--kernel",
        metavar=DOWNLOAD_KERNEL_ARGPARSE_METAVAR,
        dest="kernels",
        action="append",
        type=download_kernel_argparse_type,
        default=argparse.SUPPRESS,
        help="kernel to build for; may be given multiple times",
    )
    parser.add_argument(
        "-d",
        "--directory",
        metavar="DIR",
        type=Path,
        default="build/vmtest",
        help="directory to download assets to",
    )
    args = parser.parse_args()
    if not hasattr(args, "kernels"):
        args.kernels = []

    compilers_needed = {
        kernel.arch.name: DownloadCompiler(kernel.arch) for kernel in args.kernels
    }

    to_download = list(compilers_needed.values())
    for kernel in args.kernels:
        to_download.append(kernel)
    with download_in_thread(args.directory, to_download) as downloads:
        download_it = iter(downloads)

        for i in range(len(compilers_needed)):
            next(download_it)

        for kernel in args.kernels:
            downloaded: Kernel = next(download_it)  # type: ignore[assignment]
            print(build_kmod(args.directory, downloaded))


if __name__ == "__main__":
    _main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from collections import defaultdict
from pathlib import Path
from typing import NamedTuple

from drgn import NULL
from drgn.helpers.linux.mm import pfn_to_virt
from drgn.helpers.linux.slab import (
    find_containing_slab_cache,
    find_slab_cache,
    for_each_slab_cache,
    get_slab_cache_aliases,
    slab_cache_for_each_allocated_object,
    slab_cache_is_merged,
    slab_cache_objects_per_slab,
    slab_cache_pages_per_slab,
    slab_cache_usage,
    slab_object_info,
    slab_total_usage,
)
from tests import slow_test
from tests.linux_kernel import (
    LinuxKernelTestCase,
    meminfo_field_in_pages,
    skip_unless_have_full_mm_support,
    skip_unless_have_test_kmod,
)

SLAB_SYSFS_PATH = Path("/sys/kernel/slab")


class Slabinfo(NamedTuple):
    name: str
    active_objs: int
    num_objs: int
    objsize: int
    objperslab: int
    pagesperslab: int
    limit: int
    batchcount: int
    sharedfactor: int
    active_slabs: int
    num_slabs: int
    sharedavail: int


def iter_slabinfo():
    with open("/proc/slabinfo", "r") as f:
        # Skip the version and header.
        f.readline()
        f.readline()
        for line in f:
            components = line.split(":")
            statistics = components[0].split()
            tunables = components[1].split()
            assert tunables[0] == "tunables"
            slabdata = components[2].split()
            assert slabdata[0] == "slabdata"
            yield Slabinfo(
                name=statistics[0],
                active_objs=int(statistics[1]),
                num_objs=int(statistics[2]),
                objsize=int(statistics[3]),
                objperslab=int(statistics[4]),
                pagesperslab=int(statistics[5]),
                limit=int(tunables[1]),
                batchcount=int(tunables[2]),
                sharedfactor=int(tunables[3]),
                active_slabs=int(slabdata[1]),
                num_slabs=int(slabdata[2]),
                sharedavail=int(slabdata[3]),
            )


def get_proc_slabinfo_names():
    with open("/proc/slabinfo", "rb") as f:
        # Skip the version and header.
        f.readline()
        f.readline()
        return [line.split()[0] for line in f]


_FALLBACK_SLAB_CACHE = "mnt_cache"


def fallback_slab_cache_names(prog):
    # SLOB does not provide /proc/slabinfo. It is also disabled for SLUB if
    # CONFIG_SLUB_DEBUG=n. Before Linux kernel commit 5b36577109be ("mm:
    # slabinfo: remove CONFIG_SLABINFO") (in v4.15), it could also be disabled
    # for SLAB. So, pick a few slab caches which we know exist to test against.
    # In case they were merged into other caches, get their names from the
    # structs rather than just returning the names.
    return {
        prog[_FALLBACK_SLAB_CACHE].name.string_(),
        prog["mm_cachep"].name.string_(),
        prog["uid_cachep"].name.string_(),
    }


class TestSlab(LinuxKernelTestCase):
    @skip_unless_have_test_kmod
    def test_slab_cache_objects_per_slab(self):
        if self.prog["drgn_test_slob"]:
            self.assertRaisesRegex(
                ValueError,
                "SLOB is not supported",
                slab_cache_objects_per_slab,
                self.prog[_FALLBACK_SLAB_CACHE],
            )
            return

        try:
            for slabinfo in iter_slabinfo():
                if slabinfo.objperslab > 1:
                    # Prefer testing a slab cache with more than one object per
                    # slab.
                    break
        except FileNotFoundError:
            self.skipTest("/proc/slabinfo does not exist")
        self.assertEqual(
            slab_cache_objects_per_slab(find_slab_cache(self.prog, slabinfo.name)),
            slabinfo.objperslab,
        )

    @skip_unless_have_test_kmod
    def test_slab_cache_pages_per_slab(self):
        if self.prog["drgn_test_slob"]:
            self.assertRaisesRegex(
                ValueError,
                "SLOB is not supported",
                slab_cache_pages_per_slab,
                self.prog[_FALLBACK_SLAB_CACHE],
            )
            return

        try:
            for slabinfo in iter_slabinfo():
                if slabinfo.pagesperslab > 1:
                    # Prefer testing a slab cache with more than one page per
                    # slab.
                    break
        except FileNotFoundError:
            self.skipTest("/proc/slabinfo does not exist")
        self.assertEqual(
            slab_cache_pages_per_slab(find_slab_cache(self.prog, slabinfo.name)),
            slabinfo.pagesperslab,
        )

    def _slab_cache_aliases(self):
        if not SLAB_SYSFS_PATH.exists():
            self.skipTest(f"{str(SLAB_SYSFS_PATH)} does not exist")
        aliases = defaultdict(list)
        for child in SLAB_SYSFS_PATH.iterdir():
            if not child.name.startswith(":"):
                aliases[child.stat().st_ino].append(child.name)
        return aliases

    def test_slab_cache_is_merged_false(self):
        for aliases in self._slab_cache_aliases().values():
            if len(aliases) == 1:
                break
        else:
            self.skipTest("no unmerged slab caches")
        self.assertFalse(slab_cache_is_merged(find_slab_cache(self.prog, aliases[0])))

    def test_slab_cache_is_merged_true(self):
        for aliases in self._slab_cache_aliases().values():
            if len(aliases) > 1:
                break
        else:
            self.skipTest("no merged slab caches")
        for alias in aliases:
            slab_cache = find_slab_cache(self.prog, alias)
            if slab_cache is not None:
                break
        else:
            self.fail("couldn't find slab cache")
        self.assertTrue(slab_cache_is_merged(slab_cache))

    def test_get_slab_cache_aliases(self):
        if not SLAB_SYSFS_PATH.exists():
            # A SLOB or SLAB kernel, or one without SYSFS. Test that the
            # helper fails as expected.
            self.assertRaisesRegex(
                LookupError, "CONFIG_SYSFS", get_slab_cache_aliases, self.prog
            )
            return
        # Otherwise, the helper should work, test functionality.
        alias_to_name = get_slab_cache_aliases(self.prog)
        for aliases in self._slab_cache_aliases().values():
            # Alias groups of size 1 are either non-mergeable slabs, or
            # mergeable slabs which haven't actually been merged. Either way,
            # they should not be present in the dictionary.
            if len(aliases) == 1:
                self.assertNotIn(aliases[0], alias_to_name)
                continue

            # Find out which cache in the group is target -- it won't be
            # included in the alias dict.
            for alias in aliases:
                if alias not in alias_to_name:
                    target_alias = alias
                    aliases.remove(alias)
                    break
            else:
                self.fail("could not find target slab cache name")

            # All aliases should map to the same name
            for alias in aliases:
                self.assertEqual(alias_to_name[alias], target_alias)
            self.assertNotIn(target_alias, alias_to_name)

    def test_for_each_slab_cache(self):
        try:
            slab_cache_names = get_proc_slabinfo_names()
        except FileNotFoundError:
            # The found names should be a superset of the fallback names.
            self.assertGreaterEqual(
                {s.name.string_() for s in for_each_slab_cache(self.prog)},
                fallback_slab_cache_names(self.prog),
            )
        else:
            self.assertCountEqual(
                [s.name.string_() for s in for_each_slab_cache(self.prog)],
                slab_cache_names,
            )

    def test_find_slab_cache(self):
        try:
            slab_cache_names = get_proc_slabinfo_names()
        except FileNotFoundError:
            slab_cache_names = fallback_slab_cache_names(self.prog)
        for name in slab_cache_names:
            slab = find_slab_cache(self.prog, name)
            self.assertEqual(name, slab.name.string_())

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_slab_cache_usage(self):
        for size in ("small", "big"):
            with self.subTest(size=size):
                cache = self.prog[f"drgn_test_{size}_kmem_cache"]
                if self.prog["drgn_test_slob"]:
                    self.assertRaisesRegex(
                        ValueError, "SLOB is not supported", slab_cache_usage, cache
                    )
                    return

                name = f"drgn_test_{size}"
                try:
                    for slabinfo in iter_slabinfo():
                        if slabinfo.name == name:
                            break
                    else:
                        self.fail(f"couldn't find {name} in slabinfo")
                except FileNotFoundError:
                    self.skipTest("/proc/slabinfo does not exist")

                usage = slab_cache_usage(cache)
                self.assertEqual(
                    usage.active_objs,
                    len(self.prog[f"drgn_test_{size}_slab_objects"]),
                )
                self.assertEqual(usage.num_objs, slabinfo.num_objs)
                self.assertEqual(usage.num_slabs, slabinfo.num_slabs)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    @slow_test
    def test_slab_cache_for_each_allocated_object(self):
        for size in ("small", "big"):
            with self.subTest(size=size):
                cache = self.prog[f"drgn_test_{size}_kmem_cache"]
                objects = self.prog[f"drgn_test_{size}_slab_objects"]
                if self.prog["drgn_test_slob"]:
                    with self.assertRaisesRegex(ValueError, "SLOB is not supported"):
                        next(
                            slab_cache_for_each_allocated_object(
                                cache, f"struct drgn_test_{size}_slab_object"
                            )
                        )
                else:
                    self.assertEqual(
                        sorted(
                            slab_cache_for_each_allocated_object(
                                cache, f"struct drgn_test_{size}_slab_object"
                            ),
                            key=lambda obj: obj.value.value_(),
                        ),
                        list(objects),
                    )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_slab_object_info(self):
        for size in ("small", "big"):
            with self.subTest(size=size):
                cache = self.prog[f"drgn_test_{size}_kmem_cache"]
                objects = self.prog[f"drgn_test_{size}_slab_objects"]
                if self.prog["drgn_test_slob"]:
                    info = slab_object_info(objects[0])
                    if size == "small":
                        self.assertFalse(info.slab_cache)
                        self.assertTrue(info.slab)
                        self.assertFalse(info.address)
                        self.assertIsNone(info.allocated)
                    else:
                        self.assertIsNone(info)
                else:
                    info = slab_object_info(objects[0])
                    self.assertEqual(info.slab_cache, cache)
                    self.assertEqual(info.address, objects[0].value_())
                    self.assertTrue(info.allocated)

                    info = slab_object_info(objects[0].value.address_of_())
                    self.assertEqual(info.slab_cache, cache)
                    self.assertEqual(info.address, objects[0].value_())
                    self.assertTrue(info.allocated)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_find_containing_slab_cache(self):
        for size in ("small", "big"):
            with self.subTest(size=size):
                cache = self.prog[f"drgn_test_{size}_kmem_cache"]
                if self.prog["drgn_test_slob"]:
                    cache = NULL(self.prog, "struct kmem_cache *")
                objects = self.prog[f"drgn_test_{size}_slab_objects"]
                for obj in objects:
                    self.assertEqual(
                        find_containing_slab_cache(self.prog, obj.value_()), cache
                    )
                    self.assertEqual(find_containing_slab_cache(obj), cache)

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_find_containing_slab_cache_invalid(self):
        start_addr = pfn_to_virt(self.prog["min_low_pfn"])
        end_addr = pfn_to_virt(self.prog["max_pfn"]) + self.prog["PAGE_SIZE"]

        self.assertEqual(
            find_containing_slab_cache(self.prog, start_addr - 1),
            NULL(self.prog, "struct kmem_cache *"),
        )
        self.assertEqual(
            find_containing_slab_cache(self.prog, end_addr),
            NULL(self.prog, "struct kmem_cache *"),
        )

        self.assertEqual(
            find_containing_slab_cache(self.prog, self.prog.symbol("jiffies").address),
            NULL(self.prog, "struct kmem_cache *"),
        )

        self.assertEqual(
            find_containing_slab_cache(self.prog, self.prog["drgn_test_va"]),
            NULL(self.prog, "struct kmem_cache *"),
        )

    def test_slab_total_usage(self):
        slab_usage = slab_total_usage(self.prog)
        self.assertAlmostEqual(
            slab_usage.reclaimable_pages,
            meminfo_field_in_pages("SReclaimable"),
            delta=1024 * 1024 * 1024,
        )
        self.assertAlmostEqual(
            slab_usage.unreclaimable_pages,
            meminfo_field_in_pages("SUnreclaim"),
            delta=1024 * 1024 * 1024,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ipaddress
import os
import socket
import tempfile

from drgn import cast
from drgn.helpers.linux.fs import fget
from drgn.helpers.linux.net import (
    _S_IFMT,
    _S_IFSOCK,
    SOCK_INODE,
    SOCKET_I,
    for_each_net,
    for_each_netdev,
    get_net_ns_by_fd,
    is_pp_page,
    neigh_table_for_each_neighbor,
    netdev_for_each_tx_queue,
    netdev_get_by_index,
    netdev_get_by_name,
    netdev_ipv4_addrs,
    netdev_ipv6_addrs,
    netdev_name,
    netdev_priv,
    sk_fullsock,
    skb_shinfo,
)
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import (
    LinuxKernelTestCase,
    create_socket,
    skip_unless_have_pyroute2,
    skip_unless_have_test_kmod,
    temp_netns,
)
from util import KernelVersion


class TestNet(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.pid = os.getpid()
        cls.task = find_task(cls.prog, cls.pid)
        with open(f"/proc/{cls.pid}/ns/net") as file:
            cls.net = get_net_ns_by_fd(cls.task, file.fileno())

    def test_sk_fullsock(self):
        with create_socket() as skt:
            file = fget(self.task, skt.fileno())
            sk = cast("struct socket *", file.private_data).sk.read_()
            self.assertTrue(sk_fullsock(sk))

    def test_for_each_netdev(self):
        self.assertCountEqual(
            [netdev_name(dev) for dev in for_each_netdev(self.net)],
            [name.encode() for _, name in socket.if_nameindex()],
        )

    @skip_unless_have_test_kmod
    def test_for_each_netdev_init_net(self):
        self.assertIn(self.prog["drgn_test_netdev"], for_each_netdev(self.prog))

    @skip_unless_have_test_kmod
    def test_netdev_name(self):
        self.assertEqual(netdev_name(self.prog["drgn_test_netdev"]), b"lo")

    def test_netdev_get_by_index(self):
        for index, name in socket.if_nameindex():
            netdev = netdev_get_by_index(self.net, index)
            self.assertEqual(netdev.name.string_().decode(), name)

    def test_netdev_get_by_name(self):
        for index, name in socket.if_nameindex():
            netdev = netdev_get_by_name(self.net, name)
            self.assertEqual(netdev.ifindex, index)

    @skip_unless_have_test_kmod
    def test_netdev_get_by_name_init_net(self):
        self.assertEqual(
            netdev_get_by_name(self.prog, "lo"), self.prog["drgn_test_netdev"]
        )

    @skip_unless_have_test_kmod
    def test_netdev_priv(self):
        self.assertEqual(
            netdev_priv(self.prog["drgn_test_netdev"]),
            self.prog["drgn_test_netdev_priv"],
        )

    def test_for_each_net(self):
        self.assertIn(self.prog["init_net"].address_of_(), for_each_net(self.prog))

    def test_get_net_ns_by_fd(self):
        for index, name in socket.if_nameindex():
            netdev = netdev_get_by_index(self.net, index)
            self.assertEqual(netdev.name.string_().decode(), name)

        with tempfile.TemporaryFile("rb") as file:
            self.assertRaisesRegex(
                ValueError,
                "not a namespace inode",
                get_net_ns_by_fd,
                self.task,
                file.fileno(),
            )

        with open(f"/proc/{self.pid}/ns/mnt") as file:
            self.assertRaisesRegex(
                ValueError,
                "not a network namespace inode",
                get_net_ns_by_fd,
                self.task,
                file.fileno(),
            )

    def test_netdev_for_each_tx_queue(self):
        for index, _ in socket.if_nameindex():
            netdev = netdev_get_by_index(self.net, index)
            for queue in netdev_for_each_tx_queue(netdev):
                self.assertEqual(queue.dev, netdev)

    def test_SOCKET_I(self):
        with create_socket(type=socket.SOCK_DGRAM) as skt:
            sock = SOCKET_I(fget(self.task, skt.fileno()).f_inode)
            self.assertEqual(sock.type, socket.SOCK_DGRAM)

        with open("/dev/null") as null:
            file = fget(self.task, null.fileno())
            self.assertRaisesRegex(
                ValueError, "not a socket inode", SOCKET_I, file.f_inode
            )

    def test_SOCK_INODE(self):
        with create_socket() as skt:
            sock = SOCKET_I(fget(self.task, skt.fileno()).f_inode)
            inode = SOCK_INODE(sock)
            self.assertEqual(inode.i_mode & _S_IFMT, _S_IFSOCK)

    @skip_unless_have_test_kmod
    def test_skb_shinfo(self):
        self.assertEqual(
            skb_shinfo(self.prog["drgn_test_skb"]), self.prog["drgn_test_skb_shinfo"]
        )

    @skip_unless_have_test_kmod
    def test_is_pp_page(self):
        if not self.prog["drgn_test_have_page_pool"]:
            self.skipTest("kernel does not support page pools")
        if KernelVersion(os.uname().release) < KernelVersion("5.14"):
            self.skipTest("kernel does not support identifying page_pool pages")
        self.assertTrue(is_pp_page(self.prog["drgn_test_page_pool_page"]))
        self.assertFalse(is_pp_page(self.prog["drgn_test_page"]))

    @skip_unless_have_pyroute2
    def test_netdev_ipv4_addrs(self):
        import pyroute2

        with pyroute2.IPRoute() as ipr:
            idx = ipr.link_lookup(ifname="lo")[0]
            addrs = ipr.get_addr(index=idx, family=socket.AF_INET)
            expected = [
                ipaddress.IPv4Address(addr.get_attr("IFA_ADDRESS")) for addr in addrs
            ]
        self.assertCountEqual(
            netdev_ipv4_addrs(netdev_get_by_name(self.prog, "lo")),
            expected,
        )

    @skip_unless_have_pyroute2
    def test_netdev_ipv6_addrs(self):
        import pyroute2

        with pyroute2.IPRoute() as ipr:
            idx = ipr.link_lookup(ifname="lo")[0]
            addrs = ipr.get_addr(index=idx, family=socket.AF_INET6)
            expected = [
                ipaddress.IPv6Address(addr.get_attr("IFA_ADDRESS")) for addr in addrs
            ]
        self.assertCountEqual(
            netdev_ipv6_addrs(netdev_get_by_name(self.prog, "lo")),
            expected,
        )

    @skip_unless_have_pyroute2
    def test_neigh_table_for_each_neighbor(self):
        import pyroute2
        import pyroute2.netlink.rtnl.ndmsg

        with temp_netns() as (_, ns):
            try:
                ns.link("add", ifname="dummy0", kind="dummy")
            except pyroute2.NetlinkError:
                self.skipTest("kernel does not support dummy interface (CONFIG_DUMMY)")

            try:
                idx = ns.link_lookup(ifname="dummy0")[0]
                # Add a permanent neighbor entry so there's something to find
                ns.neigh(
                    "add",
                    dst="192.0.2.1",
                    lladdr="00:11:22:33:44:55",
                    ifindex=idx,
                    state=pyroute2.netlink.rtnl.ndmsg.states["permanent"],
                )
                nht = self.prog["arp_tbl"].nht.read_()
                neighbor_addrs = [
                    ipaddress.IPv4Address(self.prog.read(n.primary_key.address_, 4))
                    for n in neigh_table_for_each_neighbor(nht)
                ]
                self.assertIn(ipaddress.IPv4Address("192.0.2.1"), neighbor_addrs)
            finally:
                ns.link("delete", ifname="dummy0")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import unittest

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn import Object, Program, ProgramFlags
from drgn.helpers.linux.pid import find_task
from drgn.helpers.linux.stack import StackKind, kernel_stack_trace
from drgn.helpers.linux.timekeeping import ktime_get_real_seconds
from tests import TestCase
from tests.linux_kernel import skip_unless_have_stack_tracing
from tests.linux_kernel.vmcore import VMCORE_PATH, LinuxVMCoreTestCase
from util import KernelVersion


class TestVMCore(LinuxVMCoreTestCase):
    # When CPU 0 panics on s390x, the kernel switches to a different stack that
    # we don't know how to unwind through. This is a bug. For now,
    # vmtest.enter_kdump tries to avoid this so we can get some test coverage,
    # and we skip the tests otherwise.
    def _skip_if_cpu0_on_s390x(self):
        if NORMALIZED_MACHINE_NAME == "s390x" and self.prog["panic_cpu"].counter == 0:
            self.skipTest("drgn can't unwind s390x panic stack on CPU 0")

    def test_program_flags(self):
        self.assertFalse(self.prog.flags & ProgramFlags.IS_LIVE)
        self.assertTrue(self.prog.flags & ProgramFlags.IS_LINUX_KERNEL)

    def test_threads(self):
        tids = [thread.tid for thread in self.prog.threads()]
        self.assertIn(1, tids)
        self.assertIn(self.prog.crashed_thread().tid, tids)

    def test_thread(self):
        thread = self.prog.thread(1)
        self.assertEqual(thread.tid, 1)
        self.assertEqual(thread.object, find_task(self.prog, 1))

        crashed_thread_tid = self.prog.crashed_thread().tid
        self.assertEqual(self.prog.thread(crashed_thread_tid).tid, crashed_thread_tid)

    def test_thread_not_found(self):
        tids = {thread.tid for thread in self.prog.threads()}
        tid = 1
        while tid in tids:
            tid += 1
        self.assertRaises(LookupError, self.prog.thread, tid)

    def test_main_thread(self):
        self.assertRaisesRegex(
            ValueError,
            "main thread is not defined for the Linux kernel",
            self.prog.main_thread,
        )

    def test_crashed_thread(self):
        crashed_thread = self.prog.crashed_thread()
        # This assumes that we crashed from vmtest.enter_kdump. I don't know
        # why anyone would run these tests from kdump otherwise.
        self.assertEqual(crashed_thread.object.comm.string_(), b"selfdestruct")

    def _test_crashed_thread_stack_trace(self, trace):
        # This assumes that we crashed using the drgn_test kmod. Note that on
        # supported architectures, drgn_test_crash_func() is called from an IRQ
        # handler that interrupts drgn_test_crash_store().
        trace_iter = iter(trace)
        for frame in trace_iter:
            if frame.name == "drgn_test_crash_func":
                break
        else:
            self.fail("drgn_test_crash_func frame not found")

        for frame in trace_iter:
            if frame.name == "drgn_test_crash_store":
                break
        else:
            self.fail(
                "drgn_test_crash_store frame not found below drgn_test_crash_func"
            )

    @skip_unless_have_stack_tracing
    def test_crashed_thread_stack_trace(self):
        self._skip_if_cpu0_on_s390x()
        self._test_crashed_thread_stack_trace(self.prog.crashed_thread().stack_trace())

    @skip_unless_have_stack_tracing
    def test_crashed_thread_stack_trace_by_tid(self):
        self._skip_if_cpu0_on_s390x()
        self._test_crashed_thread_stack_trace(
            self.prog.stack_trace(self.prog.crashed_thread().tid)
        )

    @skip_unless_have_stack_tracing
    def test_crashed_thread_stack_trace_by_task_struct(self):
        self._skip_if_cpu0_on_s390x()
        self._test_crashed_thread_stack_trace(
            self.prog.stack_trace(self.prog.crashed_thread().object)
        )

    @skip_unless_have_stack_tracing
    def test_kernel_stack_trace(self):
        self._skip_if_cpu0_on_s390x()
        trace = kernel_stack_trace(self.prog.crashed_thread().object)
        self.assertEqual(trace.prog, self.prog)
        if NORMALIZED_MACHINE_NAME == "x86_64":
            # Since v5.1 we can identify NMI. Older kernels we just classify as
            # "exception".
            self.assertIn(trace.segments[0].kind, (StackKind.NMI, StackKind.EXCEPTION))
            # Prior to commit 946c191161cef ("x86/entry/unwind: Create stack
            # frames for saved interrupt registers"), and without ORC, drgn
            # cannot find the saved pt_regs, so the interrupted frame is not
            # marked interrupted. The full stack is still unwound, but without
            # the interrupted frame, we can't split it into multiple segments.
            if KernelVersion(os.uname().release) >= KernelVersion("4.14"):
                self.assertEqual(trace.segments[1].kind, StackKind.IRQ)
                self.assertEqual(trace.segments[2].kind, StackKind.TASK)
        elif NORMALIZED_MACHINE_NAME in ("aarch64", "arm"):
            # arm and aarch64 can unwind through IRQ stacks and correctly
            # identify them
            self.assertEqual(trace.segments[0].kind, StackKind.IRQ)
            self.assertEqual(trace.segments[1].kind, StackKind.TASK)
        else:
            # On ppc64 drgn can unwind through IRQ stacks, but there is no
            # separate IRQ stack, nor does drgn see an interrupted stack frame.
            # drgn cannot yet unwind through s390x exceptions, so we don't crash
            # from IRQ there.
            self.assertEqual(trace.segments[0].kind, StackKind.TASK)


@unittest.skipUnless(VMCORE_PATH.exists(), "not running in kdump")
class TestVMCoreNoDebugInfo(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.prog = Program()
        cls.prog.set_core_dump(VMCORE_PATH)

    @classmethod
    def tearDownClass(cls):
        del cls.prog

    def test_ktime_get_real_seconds(self):
        self.assertIsInstance(ktime_get_real_seconds(self.prog), Object)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

REGISTERS = [
    *[DrgnRegister(f"x{i}") for i in range(29)],
    DrgnRegister(["x29", "fp"]),
    DrgnRegister(["x30", "lr"]),
    DrgnRegister("sp"),
    DrgnRegister("pstate"),
]

REGISTER_LAYOUT = [
    DrgnRegisterLayout("ra_sign_state", size=8, dwarf_number=34),
    DrgnRegisterLayout("sp", size=8, dwarf_number=31),
    # Callee-saved registers.
    *[DrgnRegisterLayout(f"x{i}", size=8, dwarf_number=i) for i in range(19, 31)],
    # Caller-saved registers.
    *[DrgnRegisterLayout(f"x{i}", size=8, dwarf_number=i) for i in range(19)],
    # This pc register is only used for interrupted frames.
    DrgnRegisterLayout("pc", size=8, dwarf_number=32),
    DrgnRegisterLayout("pstate", size=8, dwarf_number=None),
]

STACK_POINTER_REGISTER = "sp"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import os.path
import tempfile

from drgn import Program
from tests import TestCase
from tests.resources import get_resource


class TestLive(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.prog = Program()
        cls.prog.set_pid(os.getpid())

    @classmethod
    def tearDownClass(cls):
        del cls.prog

    def test_threads(self):
        tids = [thread.tid for thread in self.prog.threads()]
        self.assertIn(os.getpid(), tids)
        for tid in tids:
            self.assertEqual(self.prog.thread(tid).tid, tid)

    def test_thread_not_found(self):
        self.assertRaises(LookupError, self.prog.thread, 1)

    def test_main_thread(self):
        self.assertEqual(self.prog.main_thread().tid, os.getpid())

    def test_crashed_thread(self):
        self.assertRaisesRegex(
            ValueError,
            "crashed thread is only defined for core dumps",
            self.prog.crashed_thread,
        )

    def test_thread_name(self):
        with open(f"/proc/{os.getpid()}/comm", "r") as f:
            comm = f.read().strip()
        self.assertEqual(self.prog.main_thread().name, comm)
        for thread in self.prog.threads():
            self.assertIsNotNone(thread.name)


class TestCoreDump(TestCase):
    TIDS = (
        2265413,
        2265414,
        2265415,
        2265416,
        2265417,
        2265418,
        2265419,
        2265420,
        2265421,
        2265422,
        2265423,
        2265424,
        2265425,
    )

    MAIN_TID = 2265413
    CRASHED_TID = 2265419

    MAIN_THREAD_NAME = "segfault_random"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.prog = Program()
        cls.prog.set_core_dump(get_resource("multithreaded.core"))

    @classmethod
    def tearDownClass(cls):
        del cls.prog

    def test_threads(self):
        self.assertSequenceEqual(
            sorted(thread.tid for thread in self.prog.threads()),
            self.TIDS,
        )

    def test_thread(self):
        for tid in self.TIDS:
            self.assertEqual(self.prog.thread(tid).tid, tid)

    def test_thread_not_found(self):
        self.assertRaises(LookupError, self.prog.thread, 99)

    def test_main_thread(self):
        self.assertEqual(self.prog.main_thread().tid, self.MAIN_TID)

    def test_crashed_thread(self):
        self.assertEqual(self.prog.crashed_thread().tid, self.CRASHED_TID)

    def test_thread_name(self):
        self.assertEqual(self.prog.main_thread().name, self.MAIN_THREAD_NAME)
        for tid in self.TIDS:
            if tid != self.MAIN_TID:
                self.assertIsNone(self.prog.thread(tid).name)


class TestCoreDumpLongName(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.prog = Program()
        with open(get_resource("crashme_static_pie.core"), "rb") as f:
            data = f.read()
        data = data.replace(b"crashme_static_\x00", b"crashme_static_p")
        with tempfile.NamedTemporaryFile("wb") as f:
            f.write(data)
            f.flush()
            cls.prog.set_core_dump(f.name)

    @classmethod
    def tearDownClass(cls):
        del cls.prog

    def test_thread_name(self):
        self.assertEqual(self.prog.main_thread().name, "crashme_static_p")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import FaultError
from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.percpu import per_cpu
from drgn.helpers.linux.timer import hrtimer_clock_base_for_each, timer_base_for_each
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestTimer(LinuxKernelTestCase):
    def test_timer_base_for_each_timer(self):
        timer = self.prog["drgn_test_timer"].address_of_()
        # We can race with the kernel modifying the list and either fault or
        # miss our timer, so retry a few times.
        for _ in range(10):
            for cpu in for_each_online_cpu(self.prog):
                for base in per_cpu(self.prog["timer_bases"], cpu):
                    try:
                        if timer in timer_base_for_each(base):
                            return
                    except FaultError:
                        pass
        self.fail("timer not found")

    def test_hrtimer_clock_base_for_each_hrtimer(self):
        hrtimer = self.prog["drgn_test_hrtimer"].address_of_()
        # We can race with the kernel modifying the rbtree and either fault or
        # miss our timer, so retry a few times.
        for _ in range(10):
            for cpu in for_each_online_cpu(self.prog):
                for clock_base in per_cpu(self.prog["hrtimer_bases"], cpu).clock_base:
                    try:
                        if hrtimer in hrtimer_clock_base_for_each(
                            clock_base.address_of_()
                        ):
                            return
                    except FaultError:
                        pass
        self.fail("hrtimer not found")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import ctypes
import gzip
import mmap
import os
from pathlib import Path
import re
import unittest

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn import Object, Symbol
from tests.linux_kernel import (
    IOCB_CMD_PREAD,
    IOCB_CMD_PWRITE,
    io_destroy,
    io_setup,
    io_submit,
    iocb,
    possible_cpus,
    skip_unless_have_test_kmod,
)
from tests.linux_kernel.crash_commands import CrashCommandTestCase
from tests.linux_kernel.helpers.test_irq import proc_irq_smp_affinity_list


class TestSys(CrashCommandTestCase):
    def test_sys(self):
        cmd = self.check_crash_command("sys")
        for field in (
            "KERNEL",
            "DUMPFILE",
            "CPUS",
            "DATE",
            "UPTIME",
            "LOAD AVERAGE",
            "TASKS",
            "NODENAME",
            "RELEASE",
            "VERSION",
            "MACHINE",
            "MEMORY",
        ):
            self.assertRegex(cmd.stdout, rf"(?m)^\s*{field}:")

        for variable in (
            "kernel",
            "taints",
            "dumpfile",
            "cpus",
            "offline_cpus",
            "timestamp",
            "uptime_",
            "load_average",
            "num_tasks",
            "nodename",
            "release",
            "version",
            "machine",
            "memory",
        ):
            self.assertIn(variable, cmd.drgn_option.globals)

    def test_sys_config(self):
        try:
            with gzip.open("/proc/config.gz", "rt") as f:
                expected = f.read()
        except FileNotFoundError:
            self.skipTest("kernel not built with CONFIG_IKCONFIG_PROC")

        cmd = self.check_crash_command("sys config")
        self.assertEqual(cmd.stdout, expected)
        self.assertIn("kconfig", cmd.drgn_option.globals)


class TestDev(CrashCommandTestCase):
    def test_no_args(self):
        cmd = self.check_crash_command("dev")
        with open("/proc/devices", "r") as f:
            for line in f:
                if line.isspace() or line in (
                    "Character devices:\n",
                    "Block devices:\n",
                ):
                    continue
                match = re.fullmatch(r"\s*([0-9]+)\s+(.*)\n", line)
                self.assertRegex(
                    cmd.stdout, rf"(?m)^\s*{match.group(1)}\s+{match.group(2)}"
                )

        for variable in ("cdev", "operations", "disk"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIn("major", cmd.drgn_option.globals)

    def test_ioports(self):
        cmd = self.check_crash_command("dev -i")
        lines = iter(cmd.stdout.splitlines())
        # Skip the header and root resource.
        next(lines)
        next(lines)
        with open("/proc/ioports", "r") as f:
            for output_line, file_line in zip(lines, f):
                output_match = re.fullmatch(
                    r"\s*[0-9a-f]+\s+([0-9a-f]+)-([0-9a-f]+)\s+(.*)", output_line
                )
                file_match = re.fullmatch(
                    r"\s*([0-9a-f]+)-([0-9a-f]+)\s*:\s*(.*)\n", file_line
                )
                self.assertEqual(
                    (
                        int(output_match.group(1), 16),
                        int(output_match.group(2), 16),
                        output_match.group(3),
                    ),
                    (
                        int(file_match.group(1), 16),
                        int(file_match.group(2), 16),
                        file_match.group(3),
                    ),
                )

        for variable in ("resource", "start", "end", "name"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

    def test_pci(self):
        cmd = self.check_crash_command("dev -p")
        for path in Path("/sys/bus/pci/devices").iterdir():
            self.assertIn(path.name, cmd.stdout)

        self.assertIsInstance(cmd.drgn_option.globals["root_bus"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["busname"], str)
        # Everything else is a local variable in a recursive function.

    @skip_unless_have_test_kmod
    def test_disks(self):
        cmd_d_while_idle = self.check_crash_command("dev -d", mode="capture")
        cmd_D_while_idle = self.check_crash_command("dev -D", mode="capture")

        with contextlib.ExitStack() as exit_stack:
            fd = os.open("/dev/drgntestb0", os.O_RDWR | os.O_DIRECT)
            exit_stack.callback(os.close, fd)

            major = os.major(os.fstat(fd).st_rdev)

            map = exit_stack.enter_context(
                mmap.mmap(-1, 2 * mmap.PAGESIZE, mmap.MAP_PRIVATE)
            )

            iocbs = (iocb * 2)()
            iocbs[0].aio_lio_opcode = IOCB_CMD_PWRITE
            iocbs[0].aio_fildes = fd
            iocbs[0].aio_buf = ctypes.addressof(ctypes.c_char.from_buffer(map))
            iocbs[0].aio_nbytes = mmap.PAGESIZE
            iocbs[0].aio_offset = mmap.PAGESIZE

            iocbs[1].aio_lio_opcode = IOCB_CMD_PREAD
            iocbs[1].aio_fildes = fd
            iocbs[1].aio_buf = ctypes.addressof(
                ctypes.c_char.from_buffer(map, mmap.PAGESIZE)
            )
            iocbs[1].aio_nbytes = mmap.PAGESIZE
            iocbs[1].aio_offset = 3 * mmap.PAGESIZE

            ctx_id = io_setup(len(iocbs))
            exit_stack.callback(io_destroy, ctx_id)

            truant = Path("/sys/block/drgntestb0/truant")
            truant.write_text("1\n")
            exit_stack.callback(truant.write_text, "0\n")

            io_submit(ctx_id, iocbs)

            cmd_d_while_busy = self.check_crash_command("dev -d")
            cmd_D_while_busy = self.check_crash_command("dev -D", mode="capture")

        self.assertIn("drgntestb0", cmd_d_while_idle.stdout)
        self.assertNotIn("drgntestb0", cmd_D_while_idle.stdout)
        for cmd in (cmd_d_while_busy, cmd_D_while_busy):
            self.assertRegex(
                cmd.stdout,
                rf"(?m)^\s*{major}\s+[0-9a-f]+\s+drgntestb0\s+[0-9a-f]+\s+2\s+1\s+1\s*$",
            )

        for cmd in (cmd_d_while_idle, cmd_D_while_idle, cmd_D_while_busy):
            self.assertEqual(
                cmd.drgn_option.stdout, cmd_d_while_busy.drgn_option.stdout
            )

        cmd = cmd_d_while_busy
        for variable in ("disk", "major", "queue"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIsInstance(cmd.drgn_option.globals["name"], bytes)
        for variable in ("read", "write", "total"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], int)


def nr_irqs():
    with open("/proc/stat", "r") as f:
        for line in f:
            if line.startswith("intr "):
                return len(line.split()) - 2


def nr_softirqs():
    with open("/proc/stat", "r") as f:
        for line in f:
            if line.startswith("softirq "):
                return len(line.split()) - 2


class TestIrq(CrashCommandTestCase):
    def _test_irq_common(self, cmd):
        for path in Path("/sys/kernel/irq").iterdir():
            expected = (path / "actions").read_text().rstrip("\n")
            if expected:
                irq = int(path.name)
                break
        else:
            self.skipTest("IRQ with action not found")

        action = expected.partition(",")[0]
        self.assertRegex(cmd.stdout, rf'(?m)^\s*{irq}.*"{re.escape(action)}"')

        self.assertIn("for_each_irq_desc()", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["irq"], int)
        for variable in ("irq_desc", "action", "name"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

    def test_irq(self):
        cmd = self.check_crash_command("irq")

        irqs = [int(x) for x in re.findall(r"^\s*([0-9]+)", cmd.stdout, flags=re.M)]
        self.assertEqual(irqs, list(range(nr_irqs())))

        self._test_irq_common(cmd)

    def test_used(self):
        cmd = self.check_crash_command("irq -u")

        irqs = [int(x) for x in re.findall(r"^\s*([0-9]+)", cmd.stdout, flags=re.M)]
        expected = [int(path.name) for path in Path("/sys/kernel/irq").iterdir()]
        expected.sort()
        self.assertEqual(irqs, expected)

        self._test_irq_common(cmd)

    def test_number(self):
        numbers = []
        for path in Path("/sys/kernel/irq").iterdir():
            numbers.append(int(path.name))
            if len(numbers) >= 2:
                break
        numbers.reverse()

        cmd = self.check_crash_command(f"irq {' '.join([str(x) for x in numbers])}")

        irqs = [int(x) for x in re.findall(r"^\s*([0-9]+)", cmd.stdout, flags=re.M)]
        self.assertEqual(irqs, numbers)

        self.assertIn("irq_to_desc(", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["irq"], int)
        # Omit "name" since the IRQs we chose may not have any actions with
        # names.
        for variable in ("irq_desc", "action"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

    def test_invalid_number(self):
        number = nr_irqs()
        # The drgn code would fault on a NULL irq_desc, so just compile it.
        cmd = self.check_crash_command(f"irq {number}", mode="compile")
        self.assertEqual(cmd.stdout, f"irq: invalid IRQ number: {number}\n")

    @unittest.skipUnless(NORMALIZED_MACHINE_NAME == "x86_64", "machine is not x86_64")
    def test_idt(self):
        cmd = self.check_crash_command("irq -d")

        vecs = [int(x) for x in re.findall(r"^\s*\[([0-9]+)\]", cmd.stdout, flags=re.M)]
        self.assertEqual(vecs, list(range(256)))

        self.assertRegex(cmd.stdout, r"(?m)^\s*\[3\] .*int3")

        self.assertIsInstance(cmd.drgn_option.globals["vec"], int)
        for variable in ("gate", "func"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)

    def test_softirqs(self):
        cmd = self.check_crash_command("irq -b")

        vecs = [int(x) for x in re.findall(r"^\s*\[([0-9]+)\]", cmd.stdout, flags=re.M)]
        self.assertEqual(vecs, list(range(nr_softirqs())))

        self.assertRegex(cmd.stdout, r"(?m)^\s*\[[0-9]+\] .*<\w+\>\s*$")

        self.assertIsInstance(cmd.drgn_option.globals["vec"], int)
        self.assertIsInstance(cmd.drgn_option.globals["action"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["sym"], Symbol)

    def test_affinity(self):
        cmd = self.check_crash_command("irq -a")

        proc_path = Path("/proc/irq")
        expected = [
            (
                int(path.name),
                proc_irq_smp_affinity_list(proc_path / path.name).rstrip("\n"),
            )
            for path in Path("/sys/kernel/irq").iterdir()
            if (path / "actions").read_text().rstrip("\n")
        ]
        expected.sort()
        actual = [
            (int(irq), affinity)
            for irq, affinity in re.findall(
                r"^\s*([0-9]+).*\s(\S+)\s*$", cmd.stdout, flags=re.M
            )
        ]
        self.assertEqual(actual, expected)

        self.assertIsInstance(cmd.drgn_option.globals["irq"], int)
        for variable in ("irq_desc", "affinity"):
            with self.subTest(variable=variable):
                self.assertIsInstance(cmd.drgn_option.globals[variable], Object)
        self.assertIsInstance(cmd.drgn_option.globals["names"], list)
        self.assertIsInstance(cmd.drgn_option.globals["affinity_list"], str)

    def test_stats(self):
        cmd = self.check_crash_command("irq -s")

        expected = [
            int(path.name)
            for path in Path("/sys/kernel/irq").iterdir()
            if (path / "actions").read_text().rstrip("\n")
        ]
        expected.sort()
        num_possible_cpus = len(possible_cpus())
        actual = [
            int(irq)
            for irq in re.findall(
                rf"^\s*([0-9]+):(?:\s+[0-9]+){{{num_possible_cpus}}}",
                cmd.stdout,
                flags=re.M,
            )
        ]
        self.assertEqual(actual, expected)

        self.assertIsInstance(cmd.drgn_option.globals["irq"], int)
        self.assertIsInstance(cmd.drgn_option.globals["irq_desc"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["names"], list)
        self.assertIn("chip_name", cmd.drgn_option.globals)
        self.assertIsInstance(cmd.drgn_option.globals["count"], int)

    def test_stats_cpu(self):
        cpu = max(possible_cpus())
        cmd = self.check_crash_command(f"irq -s -c {cpu}")

        expected = [
            int(path.name)
            for path in Path("/sys/kernel/irq").iterdir()
            if (path / "actions").read_text().rstrip("\n")
        ]
        expected.sort()
        actual = [
            int(irq)
            for irq in re.findall(r"^\s*([0-9]+):\s+[0-9]+", cmd.stdout, flags=re.M)
        ]
        self.assertEqual(actual, expected)

        self.assertEqual(cmd.drgn_option.globals["cpu"], cpu)
        self.assertIsInstance(cmd.drgn_option.globals["irq"], int)
        self.assertIsInstance(cmd.drgn_option.globals["irq_desc"], Object)
        self.assertIsInstance(cmd.drgn_option.globals["names"], list)
        self.assertIn("chip_name", cmd.drgn_option.globals)
        self.assertIsInstance(cmd.drgn_option.globals["count"], int)


class TestLog(CrashCommandTestCase):
    def test_no_args(self):
        cmd = self.check_crash_command("log")

        self.assertRegex(cmd.stdout, r"^\[\s*[0-9]+\.[0-9]+\] .")

        self.assertIn("get_dmesg()", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["dmesg"], bytes)

    def test_t(self):
        cmd = self.check_crash_command("log -t")

        self.assertRegex(cmd.stdout, r"^[^[]")

        self.assertIn("get_dmesg(timestamps=False)", cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["dmesg"], bytes)

    def test_T(self):
        cmd = self.check_crash_command("log -T")

        self.assertRegex(cmd.stdout, r"^\[[A-Z]")

        self.assertIn('get_dmesg(timestamps="human")', cmd.drgn_option.stdout)
        self.assertIsInstance(cmd.drgn_option.globals["dmesg"], bytes)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import logging
from pathlib import Path
import subprocess
import tempfile
from typing import Literal, Optional, TextIO

from vmtest.chroot import chroot_sh_cmd
from vmtest.config import (
    ARCHITECTURES,
    HOST_ARCHITECTURE,
    Architecture,
    _run_autoreconf,
)

logger = logging.getLogger(__name__)


_ROOTFS_PACKAGES = (
    # drgn build dependencies.
    "autoconf",
    "automake",
    "gcc",
    "git",
    "libdw-dev",
    "libelf-dev",
    "libjson-c-dev",
    "libkdumpfile-dev",
    "liblzma-dev",
    "libpcre2-dev",
    "libtool",
    "make",
    "pkgconf",
    "python3",
    "python3-dev",
    "python3-pip",
    "python3-setuptools",
    # Test dependencies.
    "btrfs-progs",
    "check",
    "e2fsprogs",
    "iproute2",
    "kexec-tools",
    "kmod",
    "python3-pyroute2",
    "python3-pytest",
    "python3-pytest-subtests",
    "zstd",
)


def build_rootfs(
    arch: Architecture,
    path: Path,
    *,
    btrfs: Literal["never", "always", "auto"] = "auto",
) -> None:
    if path.exists():
        logger.info("%s already exists", path)
        return

    packages = list(_ROOTFS_PACKAGES)
    if arch is HOST_ARCHITECTURE:
        for other_arch in ARCHITECTURES.values():
            if other_arch is not arch:
                packages.append(f"gcc-{other_arch.debian_gcc_target}")

    # As of Debian 13, kexec-tools was not available for RISC-V:
    # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=1019254. We can install
    # it from testing for now.
    if arch.name == "riscv64":
        packages.remove("kexec-tools")
        riscv_kexec_tools_workaround = r"""
cat > /etc/apt/sources.list.d/testing.list << "EOF"
deb https://deb.debian.org/debian testing main
EOF

cat > /etc/apt/preferences.d/99testing << "EOF"
Package: *
Pin: release a=testing
Pin-Priority: 100
EOF

apt update
apt install -y -t testing kexec-tools
"""
    else:
        riscv_kexec_tools_workaround = ""

    logger.info("creating debootstrap rootfs %s", path)
    path.parent.mkdir(parents=True, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=path.parent) as tmp_name:
        tmp_dir = Path(tmp_name)
        snapshot = False

        if btrfs != "never":
            try:
                import btrfsutil  # type: ignore  # No type hints available.

                btrfsutil.create_subvolume(tmp_dir / path.name)
                snapshot = True
            except (ImportError, OSError):
                if btrfs == "always":
                    raise

        subprocess.check_call(
            [
                "unshare",
                "--map-root-user",
                "--map-auto",
                "sh",
                "-c",
                rf"""
set -e

arch="$1"
target="$2"
packages="$3"

# We're not really an LXC container, but this convinces debootstrap to skip
# some operations that it can't do in a user namespace.
container=lxc debootstrap --variant=minbase --foreign --include="$packages" --arch="$arch" stable "$target"
# By default, apt drops privileges to the "_apt" user. unshare --map-auto can
# deal with that, but it creates files that can't be accessed as the normal
# user without unshare, which is annoying.
echo 'APT::Sandbox::User "root";' > "$target/etc/apt/apt.conf.d/99nosandbox"
{chroot_sh_cmd('"$target"')} '
set -e

container=lxc /debootstrap/debootstrap --second-stage
{riscv_kexec_tools_workaround}
apt clean'
""",
                "sh",
                arch.debian_arch,
                tmp_dir / path.name,
                ",".join(packages),
            ]
        )
        (tmp_dir / path.name).rename(path)
    logger.info("created debootstrap rootfs %s", path)

    if snapshot:
        snapshot_dir = path.parent / (path.name + ".pristine")
        btrfsutil.create_snapshot(path, snapshot_dir, read_only=True)
        logger.info("created snapshot %s", snapshot_dir)


def _cross_compile_drgn(
    target: Architecture, directory: Path, *, outfile: Optional[TextIO] = None
) -> None:
    assert HOST_ARCHITECTURE is not None
    host_rootfs = directory / HOST_ARCHITECTURE.name / "rootfs"
    target_rootfs = directory / target.name / "rootfs"
    logger.info("cross-compiling drgn in %s for %s", host_rootfs, target_rootfs)
    subprocess.check_call(
        [
            "unshare",
            "--map-root-user",
            "--map-auto",
            "--mount",
            "sh",
            "-c",
            rf"""
set -e

mount --bind . "$1/mnt"
mount --bind "$3" "$1/srv"
{chroot_sh_cmd('"$1"')} '
set -e

extension_suffix="$(/srv/usr/bin/python3-config --extension-suffix)"
build_temp="/mnt/build/temp${{extension_suffix%.so}}"
mkdir -p "$build_temp"
cd "$build_temp"
export CFLAGS="--sysroot=/srv"
export LDFLAGS="--sysroot=/srv"
export PYTHON_CPPFLAGS="$(/srv/usr/bin/python3-config --includes)"
export PKG_CONFIG_LIBDIR="$(pkg-config --variable pc_path pkg-config | sed -r -e "s!(^|:)/!\1/srv/!g" -e "s/$1/$2/g")"
export PKG_CONFIG_SYSROOT_DIR=/srv
if [ ! -e Makefile ]; then
    /mnt/libdrgn/configure --host="$2" --with-sysroot=/srv --disable-static --disable-libdrgn --enable-python-extension
fi
make -j"$(nproc)"
cp -v ".libs/_drgn.so" "/mnt/_drgn$extension_suffix"
' sh "$2" "$4"
""",
            "sh",
            host_rootfs,
            HOST_ARCHITECTURE.debian_gcc_target,
            target_rootfs,
            target.debian_gcc_target,
        ],
        stdout=outfile,
        stderr=outfile,
    )


def _build_drgn_in_rootfs(rootfs: Path, *, outfile: Optional[TextIO] = None) -> None:
    logger.info("building drgn using %s", rootfs)
    subprocess.check_call(
        [
            "unshare",
            "--map-root-user",
            "--map-auto",
            "--mount",
            "sh",
            "-c",
            rf"""
set -e

mount --bind . "$1/mnt"
{chroot_sh_cmd('"$1"')} 'cd /mnt && CONFIGURE_FLAGS=--enable-compiler-warnings=error python3 setup.py build_ext -i'
""",
            "sh",
            rootfs,
        ],
        stdout=outfile,
        stderr=outfile,
    )


def build_drgn_for_arch(
    target: Architecture, directory: Path, *, outfile: Optional[TextIO] = None
) -> None:
    _run_autoreconf()
    if target is not HOST_ARCHITECTURE and HOST_ARCHITECTURE is not None:
        return _cross_compile_drgn(target, directory, outfile=outfile)
    else:
        return _build_drgn_in_rootfs(
            directory / target.name / "rootfs", outfile=outfile
        )


if __name__ == "__main__":
    import argparse

    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", level=logging.INFO
    )

    parser = argparse.ArgumentParser(
        description="build root filesystems for vmtest. "
        "This requires debootstrap(8), qemu-user-static, and unprivileged user namespaces.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--directory",
        metavar="DIR",
        type=Path,
        default="build/vmtest",
        help="directory for vmtest artifacts",
    )
    parser.add_argument(
        "--build-drgn",
        action="store_true",
        help="also build drgn in the current directory using the built rootfs",
    )
    parser.add_argument(
        "--btrfs",
        choices=["never", "always", "auto"],
        default="auto",
        help="make the rootfs a Btrfs subvolume and create a read-only snapshot",
    )
    parser.add_argument(
        "-a",
        "--architecture",
        dest="architectures",
        action="append",
        choices=["all", "foreign", *sorted(ARCHITECTURES)],
        default=argparse.SUPPRESS,
        help='architecture to build for, or "foreign" for all architectures other than the host architecture; may be given multiple times (default: foreign)',
    )
    args = parser.parse_args()

    if not hasattr(args, "architectures"):
        args.architectures = ["foreign"]
    architectures = []
    for name in args.architectures:
        if name == "foreign":
            architectures.extend(
                [
                    arch
                    for arch in ARCHITECTURES.values()
                    if arch is not HOST_ARCHITECTURE
                ]
            )
        elif name == "all":
            architectures.extend(ARCHITECTURES.values())
        else:
            architectures.append(ARCHITECTURES[name])

    # Create the host architecture rootfs first in case it's needed for
    # building drgn for other architectures.
    try:
        index = architectures.index(HOST_ARCHITECTURE)  # type: ignore[arg-type]
    except ValueError:
        pass
    else:
        architectures.insert(0, architectures.pop(index))

    for arch in architectures:
        dir = args.directory / arch.name / "rootfs"
        build_rootfs(arch, dir, btrfs=args.btrfs)
        if args.build_drgn:
            build_drgn_for_arch(arch, args.directory)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import shlex

from tests.linux_kernel import skip_unless_have_test_kmod
from tests.linux_kernel.crash_commands import CrashCommandTestCase


@skip_unless_have_test_kmod
class TestMod(CrashCommandTestCase):
    def test_no_options(self):
        cmd = self.check_crash_command("mod")
        self.assertIn("drgn_test", cmd.stdout)
        self.assertEqual(
            cmd.drgn_option.globals["struct_module"].type_.type_name(),
            "struct module *",
        )

    def test_taints(self):
        cmd = self.check_crash_command("mod -t")
        self.assertRegex(cmd.stdout, "drgn_test.*O")
        self.assertIsInstance(cmd.drgn_option.globals["taints"], str)

    def test_load(self):
        cmd = self.check_crash_command("mod -s drgn_test")
        self.assertIn("drgn_test", cmd.stdout)
        self.assertIn("prog.load_module_debug_info", cmd.drgn_option.stdout)

    def test_load_path(self):
        cmd = self.check_crash_command(
            f"mod -s drgn_test {shlex.quote(os.getenv('DRGN_TEST_KMOD'))}"
        )
        self.assertIn("drgn_test", cmd.stdout)
        self.assertIn("module.try_file", cmd.drgn_option.stdout)

    def test_load_all(self):
        cmd = self.check_crash_command("mod -S")
        self.assertIn("drgn_test", cmd.stdout)
        self.assertIn("prog.load_default_debug_info", cmd.drgn_option.stdout)

    def test_load_all_directory(self):
        cmd = self.check_crash_command("mod -S /dev/null")
        self.assertIn("drgn_test", cmd.stdout)
        self.assertIn(
            "prog.debug_info_options.kernel_directories", cmd.drgn_option.stdout
        )
        self.assertIn("prog.load_default_debug_info", cmd.drgn_option.stdout)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("max", type=int)
    args = parser.parse_args()

    print(
        "#define PP_MAP(func, arg, ...) PP_OVERLOAD(PP_MAP_I, __VA_ARGS__)(func, arg, __VA_ARGS__)"
    )
    print("/** @cond */")
    for i in range(args.max, 1, -1):
        print(
            f"#define PP_MAP_I{i}(func, arg, x, ...) func(arg, x) PP_MAP_I{i - 1}(func, arg, __VA_ARGS__)"
        )
    print("#define PP_MAP_I1(func, arg, x) func(arg, x)")
    print("#define PP_MAP_I0(func, arg, x)")
    print("/** @endcond */")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path

from drgn.helpers.linux.pci import (
    for_each_pci_dev,
    for_each_pci_root_bus,
    pci_bus_for_each_child,
    pci_bus_for_each_dev,
    pci_bus_name,
    pci_is_bridge,
    pci_name,
    pci_pcie_type,
)
from tests.linux_kernel import LinuxKernelTestCase


def _pci_root_bus_paths():
    for path in Path("/sys/class/pci_bus").iterdir():
        real_device_path = (path / "device").resolve()
        # For root buses, the device name is "pciDDDD:BB". For other buses, it
        # is the bridge device name (DDDD:BB:dd.F).
        if real_device_path.name.startswith("pci"):
            yield real_device_path


def _pcie_type(config_path) -> int:
    with open(config_path, "rb") as f:
        f.seek(0x34)
        cap_ptr = f.read(1)[0]

        while cap_ptr and cap_ptr < 0xFF:
            f.seek(cap_ptr)
            cap_id = f.read(1)[0]

            if cap_id == 0x10:
                f.seek(cap_ptr + 2)
                pcie_caps = int.from_bytes(f.read(2), byteorder="little")
                return (pcie_caps >> 4) & 0x0F

            f.seek(cap_ptr + 1)
            cap_ptr = f.read(1)[0]

    return None


class TestPci(LinuxKernelTestCase):
    def test_for_each_pci_dev(self):
        # This also tests pci_name().
        self.assertCountEqual(
            [pci_name(dev) for dev in for_each_pci_dev(self.prog)],
            [path.name for path in Path("/sys/bus/pci/devices").iterdir()],
        )

    def test_for_each_pci_root_bus(self):
        # This also tests pci_bus_name().
        self.assertCountEqual(
            [pci_bus_name(bus) for bus in for_each_pci_root_bus(self.prog)],
            [path.name[3:] for path in _pci_root_bus_paths()],
        )

    def test_pci_bus_for_each_child(self):
        bus_path = next(_pci_root_bus_paths())
        bus_name = bus_path.name[3:]
        for bus in for_each_pci_root_bus(self.prog):
            if pci_bus_name(bus) == bus_name:
                break
        else:
            self.fail(f"root bus {bus_path.name} not found")

        self.assertCountEqual(
            [pci_bus_name(child_bus) for child_bus in pci_bus_for_each_child(bus)],
            [child_bus_path.name for child_bus_path in bus_path.glob("*/pci_bus/*")],
        )

    def test_pci_bus_for_each_dev(self):
        bus_path = next(_pci_root_bus_paths())
        bus_name = bus_path.name[3:]
        for bus in for_each_pci_root_bus(self.prog):
            if pci_bus_name(bus) == bus_name:
                break
        else:
            self.fail(f"root bus {bus_path.name} not found")

        self.assertCountEqual(
            [pci_name(dev) for dev in pci_bus_for_each_dev(bus)],
            [dev_path.name for dev_path in bus_path.glob(bus_name + ":*")],
        )

    def test_pci_is_bridge(self):
        bridge_name = None
        non_bridge_name = None
        for path in Path("/sys/bus/pci/devices").iterdir():
            if (path / "secondary_bus_number").exists():
                bridge_name = path.name
                if non_bridge_name is not None:
                    break
            else:
                non_bridge_name = path.name
                if bridge_name is not None:
                    break

        bridge_dev = None
        non_bridge_dev = None
        for dev in for_each_pci_dev(self.prog):
            name = pci_name(dev)
            if name == bridge_name:
                bridge_dev = dev
                if non_bridge_name is None or non_bridge_dev is not None:
                    break
            elif name == non_bridge_name:
                non_bridge_dev = dev
                if bridge_name is None or bridge_dev is not None:
                    break

        with self.subTest("bridge"):
            if bridge_name is None:
                self.skipTest("no bridges found")
            if bridge_dev is None:
                self.fail(f"device {bridge_name} not found")
            self.assertTrue(pci_is_bridge(bridge_dev))

        with self.subTest("non-bridge"):
            if non_bridge_name is None:
                self.skipTest("no non-bridges found")
            if non_bridge_dev is None:
                self.fail(f"device {non_bridge_name} not found")
            self.assertFalse(pci_is_bridge(non_bridge_dev))

    def test_pci_pcie_type(self):
        for path in Path("/sys/bus/pci/devices").iterdir():
            expected = _pcie_type(path / "config")
            if expected is not None:
                break
        else:
            self.skipTest("no PCIe devices found")

        for dev in for_each_pci_dev(self.prog):
            if pci_name(dev) == path.name:
                break
        else:
            self.fail(f"device {path.name} not found")

        self.assertEqual(pci_pcie_type(dev), expected)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ast
import dataclasses
import inspect
import operator
import os.path
import re
import stat
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)

from drgndoc.visitor import NodeVisitor


class _PreTransformer(ast.NodeTransformer):
    # Replace string forward references with the parsed expression.
    @overload
    def _visit_annotation(self, node: ast.expr) -> ast.expr: ...

    @overload
    def _visit_annotation(self, node: None) -> None: ...

    def _visit_annotation(self, node: Optional[ast.expr]) -> Optional[ast.expr]:
        if isinstance(node, ast.Constant) and isinstance(node.value, str):
            node = self.visit(ast.parse(node.value, "<string>", "eval").body)
        return node

    def visit_arg(self, node: ast.arg) -> ast.arg:
        node = cast(ast.arg, self.generic_visit(node))
        node.annotation = self._visit_annotation(node.annotation)
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        node = cast(ast.FunctionDef, self.generic_visit(node))
        node.returns = self._visit_annotation(node.returns)
        return node

    def visit_AsyncFunctionDef(
        self, node: ast.AsyncFunctionDef
    ) -> ast.AsyncFunctionDef:
        node = cast(ast.AsyncFunctionDef, self.generic_visit(node))
        node.returns = self._visit_annotation(node.returns)
        return node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.AnnAssign:
        node = cast(ast.AnnAssign, self.generic_visit(node))
        node.annotation = self._visit_annotation(node.annotation)
        return node

    # Get rid of Index nodes, which are deprecated as of Python 3.9.
    def visit_Index(self, node: Any) -> Any:
        return self.visit(node.value)


@dataclasses.dataclass
class Module:
    path: Optional[str]
    docstring: Optional[str]
    attrs: Mapping[str, "Node"]

    def has_docstring(self) -> bool:
        return self.docstring is not None


@dataclasses.dataclass
class Class:
    bases: Sequence[ast.expr]
    docstring: Optional[str]
    attrs: Mapping[str, "NonModuleNode"]

    def has_docstring(self) -> bool:
        if self.docstring is not None:
            return True
        init = self.attrs.get("__init__")
        return isinstance(init, Function) and init.has_docstring()


@dataclasses.dataclass
class FunctionSignature:
    args: ast.arguments
    returns: Optional[ast.expr]
    decorator_list: Sequence[ast.expr]
    docstring: Optional[str]

    def has_decorator(self, name: str) -> bool:
        return any(
            isinstance(decorator, ast.Name) and decorator.id == name
            for decorator in self.decorator_list
        )


@dataclasses.dataclass
class Function:
    async_: bool
    signatures: Sequence[FunctionSignature]

    def has_docstring(self) -> bool:
        return any(signature.docstring is not None for signature in self.signatures)


@dataclasses.dataclass
class Variable:
    annotation: Optional[ast.expr]
    value: Optional[ast.expr]
    docstring: Optional[str]

    def has_docstring(self) -> bool:
        return self.docstring is not None


@dataclasses.dataclass
class Import:
    module: str
    aliased: bool

    def has_docstring(self) -> bool:
        return False


@dataclasses.dataclass
class ImportFrom:
    name: str
    module: Optional[str]
    level: int
    aliased: bool

    def has_docstring(self) -> bool:
        return False


Node = Union[Module, Class, Function, Variable, Import, ImportFrom]
NonModuleNode = Union[Class, Function, Variable, Import, ImportFrom]
DocumentedNode = Union[Module, Class, Function, Variable]


def _docstring_from_node(node: Optional[ast.AST]) -> Optional[str]:
    if not isinstance(node, ast.Expr):
        return None
    node = node.value
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        text = node.value
    else:
        return None
    return inspect.cleandoc(text)


def _transform_function(node: Function) -> Function:
    signature = node.signatures[-1]

    if (
        signature.has_decorator("takes_program_or_default")
        and signature.docstring is not None
    ):
        match = re.search(
            r"^(\s*):(?:param|return|raises)", signature.docstring, flags=re.M
        )
        if match:
            prefix = match.group(1)
            pos = match.start()
        else:
            prefix = "\n\n"
            pos = len(signature.docstring)
        signature.docstring = "".join(
            (
                signature.docstring[:pos],
                prefix,
                ":param prog: Program, which :ref:`may be omitted to use the default program argument <default-program>`.",
                signature.docstring[pos:],
            )
        )

    if signature.has_decorator("takes_object_or_program_or_default"):
        del signature.args.args[0]
        signature.args.args[0].annotation = ast.Subscript(
            value=ast.Name(id="Union", ctx=ast.Load()),
            slice=ast.Tuple(
                elts=[
                    ast.Name(id="Object", ctx=ast.Load()),
                    ast.Name(id="Program", ctx=ast.Load()),
                ],
                ctx=ast.Load(),
            ),
            ctx=ast.Load(),
        )

    return node


class _ModuleVisitor(NodeVisitor):
    def visit(self, node: ast.AST) -> Tuple[Optional[str], Dict[str, NonModuleNode]]:
        self._attrs: Dict[str, NonModuleNode] = {}
        super().visit(node)
        docstring = self._docstring
        del self._docstring
        return docstring, self._attrs

    def visit_Module(
        self, node: ast.Module, parent: Optional[ast.AST], sibling: Optional[ast.AST]
    ) -> None:
        self._docstring = ast.get_docstring(node)
        self.generic_visit(node)

    def visit_ClassDef(
        self, node: ast.ClassDef, parent: Optional[ast.AST], sibling: Optional[ast.AST]
    ) -> None:
        attrs = self._attrs
        self._attrs = {}
        self.generic_visit(node)
        class_node = Class(node.bases, ast.get_docstring(node), self._attrs)
        self._attrs = attrs
        self._attrs[node.name] = class_node

    def _visit_function(
        self,
        node: Union[ast.FunctionDef, ast.AsyncFunctionDef],
        parent: Optional[ast.AST],
        sibling: Optional[ast.AST],
    ) -> None:
        signature = FunctionSignature(
            node.args, node.returns, node.decorator_list, ast.get_docstring(node)
        )
        async_ = isinstance(node, ast.AsyncFunctionDef)
        func = self._attrs.get(node.name)
        # If we have a previous overload definition, we can add to it.
        # Otherwise, we replace it.
        if (
            func
            and isinstance(func, Function)
            and func.async_ == async_
            and func.signatures[-1].has_decorator("overload")
        ):
            signatures = list(func.signatures)
            signatures.append(signature)
        else:
            signatures = [signature]
        self._attrs[node.name] = _transform_function(Function(async_, signatures))
        # NB: we intentionally don't visit the function body.

    visit_FunctionDef = _visit_function
    visit_AsyncFunctionDef = _visit_function

    def _add_assign(
        self,
        name: str,
        annotation: Optional[ast.expr],
        value: Optional[ast.expr],
        docstring: Optional[str],
    ) -> None:
        try:
            var = self._attrs[name]
        except KeyError:
            pass
        else:
            # The name was previously defined. If it's a variable, add the
            # annotation, value, and/or docstring. If this is an annotation
            # without a value, don't do anything. Otherwise, replace the
            # previous definition.
            if isinstance(var, Variable):
                if annotation is None:
                    annotation = var.annotation
                if value is None:
                    value = var.value
                if docstring is None:
                    docstring = var.docstring
            elif value is None:
                return
        self._attrs[name] = Variable(annotation, value, docstring)

    def visit_Assign(
        self, node: ast.Assign, parent: Optional[ast.AST], sibling: Optional[ast.AST]
    ) -> None:
        if len(node.targets) == 1:
            docstring = _docstring_from_node(sibling)
        else:
            docstring = None
        for target in node.targets:
            if isinstance(target, ast.Name):
                self._add_assign(target.id, None, node.value, docstring)

    def visit_AnnAssign(
        self, node: ast.AnnAssign, parent: Optional[ast.AST], sibling: Optional[ast.AST]
    ) -> None:
        if isinstance(node.target, ast.Name):
            self._add_assign(
                node.target.id,
                node.annotation,
                node.value,
                _docstring_from_node(sibling),
            )

    def visit_Import(
        self, node: ast.Import, parent: Optional[ast.AST], sibling: Optional[ast.AST]
    ) -> None:
        for alias in node.names:
            if alias.asname is None:
                # We don't distinguish between "import foo" and "import
                # foo.bar"; they both add "foo" to the current scope.
                name = module_name = alias.name.partition(".")[0]
            else:
                name = alias.asname
                module_name = alias.name
            self._attrs[name] = Import(module_name, alias.asname is not None)

    def visit_ImportFrom(
        self,
        node: ast.ImportFrom,
        parent: Optional[ast.AST],
        sibling: Optional[ast.AST],
    ) -> None:
        for alias in node.names:
            name = alias.name if alias.asname is None else alias.asname
            self._attrs[name] = ImportFrom(
                alias.name, node.module, node.level, alias.asname is not None
            )


def parse_source(
    source: str, filename: str
) -> Tuple[Optional[str], Dict[str, NonModuleNode]]:
    node = ast.parse(source, filename)
    return _ModuleVisitor().visit(_PreTransformer().visit(node))


def _default_handle_err(e: Exception) -> None:
    raise e


def parse_module(
    path: str, handle_err: Callable[[Exception], None] = _default_handle_err
) -> Optional[Tuple[Optional[str], Dict[str, NonModuleNode]]]:
    try:
        with open(path, "r") as f:
            source = f.read()
    except (OSError, UnicodeError) as e:
        handle_err(e)
        return None
    try:
        return parse_source(source, path)
    except SyntaxError as e:
        handle_err(e)
        return None


def parse_package(
    path: str, handle_err: Callable[[Exception], None] = _default_handle_err
) -> Optional[Module]:
    module_path: Optional[str] = None
    docstring: Optional[str] = None
    attrs: Dict[str, Node] = {}
    init_path = os.path.join(path, "__init__.py")
    if os.path.isfile(init_path):
        module_path = init_path
        result = parse_module(init_path, handle_err)
        if result is not None:
            docstring = result[0]
            attrs = cast(Dict[str, Node], result[1])

    try:
        entries = sorted(os.scandir(path), key=operator.attrgetter("name"))
    except OSError as e:
        handle_err(e)
    else:
        for entry in entries:
            try:
                is_dir = entry.is_dir()
                is_file = entry.is_file()
            except OSError as e:
                handle_err(e)
                continue
            if is_dir:
                subpackage = parse_package(entry.path, handle_err)
                if subpackage:
                    attrs[entry.name] = subpackage
            elif is_file and entry.name != "__init__.py":
                root, ext = os.path.splitext(entry.name)
                if ext == ".py" or ext == ".pyi":
                    result = parse_module(entry.path, handle_err)
                    if result:
                        attrs[root] = Module(entry.path, result[0], result[1])

    if module_path is None and docstring is None and not attrs:
        return None
    return Module(module_path, docstring, attrs)


def parse_paths(
    paths: Iterable[str], handle_err: Callable[[Exception], None] = _default_handle_err
) -> Mapping[str, Module]:
    modules = {}
    for path in paths:
        path = os.path.realpath(path)
        try:
            st = os.stat(path)
        except OSError as e:
            handle_err(e)
            continue
        if stat.S_ISDIR(st.st_mode):
            package = parse_package(path, handle_err)
            if package:
                modules[os.path.basename(path)] = package
            else:
                handle_err(Exception(f"{path}:Not a Python module or package"))
        else:
            result = parse_module(path, handle_err)
            if result:
                name = os.path.splitext(os.path.basename(path))[0]
                modules[name] = Module(path, result[0], result[1])
    return modules



# Copyright (c) 2026 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

from tests.linux_kernel import (
    fork_and_stop,
    skip_unless_have_stack_tracing,
    skip_unless_have_test_kmod,
)
from tests.linux_kernel.crash_commands import CrashCommandTestCase


@skip_unless_have_stack_tracing
class TestBt(CrashCommandTestCase):
    def test_no_options(self):
        with fork_and_stop() as pid:
            self.run_crash_command(f"set {pid}")
            cmd = self.check_crash_command("bt")
            self.assertIn(f"PID: {pid}", cmd.stdout)
            self.assertRegex(
                cmd.stdout, r"(?m).*^ *#\d+ \[[0-9a-f]+\] \w+ at [0-9a-f]+$"
            )

    def test_pid(self):
        with fork_and_stop() as pid:
            cmd = self.check_crash_command(f"bt {pid}")
            self.assertIn(f"PID: {pid}", cmd.stdout)
            self.assertRegex(
                cmd.stdout, r"(?m).*^ *#\d+ \[[0-9a-f]+\] \w+ at [0-9a-f]+$"
            )

    def test_multiple_pids(self):
        with fork_and_stop() as pid1, fork_and_stop() as pid2:
            cmd = self.check_crash_command(f"bt {pid1} {pid2}")
            self.assertIn(f"PID: {pid1}", cmd.stdout)
            self.assertIn(f"PID: {pid2}", cmd.stdout)

    @skip_unless_have_test_kmod
    def test_task_struct(self):
        task = self.prog["drgn_test_kthread"]
        cmd = self.check_crash_command(f"bt 0x{task.value_():x}")
        self.assertIn(f"PID: {task.pid.value_()}", cmd.stdout)
        self.assertIn(f"TASK: {task.value_():x}", cmd.stdout)
        self.assertIn("drgn_test_kthread_fn", cmd.stdout)
        self.assertIn("drgn_test_kthread_fn2", cmd.stdout)
        self.assertIn("drgn_test_kthread_fn3", cmd.stdout)
        self.assertIn("[drgn_test]", cmd.stdout)

    @skip_unless_have_test_kmod
    def test_frame_data(self):
        task = self.prog["drgn_test_kthread"]
        cmd = self.check_crash_command(f"bt -f 0x{task.value_():x}")
        self.assertRegex(cmd.stdout, "(?m)^    [0-9a-f]+:  [0-9a-f]+")

    @skip_unless_have_test_kmod
    def test_frame_data_slab(self):
        task = self.prog["drgn_test_kthread"]
        cmd = self.check_crash_command(f"bt -F 0x{task.value_():x}")
        # We know we will find return address symbols on the stack:
        self.assertIn("drgn_test_kthread_fn+", cmd.stdout)
        self.assertIn("drgn_test_kthread_fn2+", cmd.stdout)
        self.assertIn("drgn_test_kthread_fn3+", cmd.stdout)
        # We also know we will find the drgn small slab object, but whether it
        # can be identified depends on kernel configuration (non-SLOB)
        if self.prog["drgn_test_slob"]:
            self.assertIn("[unknown slab object]", cmd.stdout)
        else:
            self.assertIn("[drgn_test_small]", cmd.stdout)

    @skip_unless_have_test_kmod
    def test_frame_data_verbose(self):
        task = self.prog["drgn_test_kthread"]
        cmd = self.check_crash_command(f"bt -FF 0x{task.value_():x}")
        # We know we will find return address symbols on the stack:
        self.assertIn("drgn_test_kthread_fn+", cmd.stdout)
        self.assertIn("drgn_test_kthread_fn2+", cmd.stdout)
        self.assertIn("drgn_test_kthread_fn3+", cmd.stdout)
        # We also know we will find the drgn small slab object, but whether it
        # can be identified depends on kernel configuration (non-SLOB)
        if self.prog["drgn_test_slob"]:
            self.assertRegex(cmd.stdout, r"\[[0-9a-f]+:unknown slab object\]")
        else:
            self.assertRegex(cmd.stdout, r"\[[0-9a-f]+:drgn_test_small\]")

    @skip_unless_have_test_kmod
    def test_line_numbers(self):
        task = self.prog["drgn_test_kthread"]
        cmd = self.check_crash_command(f"bt -l 0x{task.value_():x}")
        self.assertIn("drgn_test.c", cmd.stdout)

    @skip_unless_have_test_kmod
    def test_variables(self):
        task = self.prog["drgn_test_kthread"]
        cmd = self.check_crash_command(f"bt -V 0x{task.value_():x}")
        self.assertIn("a = (volatile int)1", cmd.stdout)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
from pathlib import Path
import random
import unittest

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn import NULL, FaultError, Object
from drgn.helpers.experimental.kmodify import (
    call_function,
    call_functions,
    clear_bit,
    pass_pointer,
    set_bit,
    write_memory,
    write_object,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod

skip_unless_have_kmodify = unittest.skipUnless(
    NORMALIZED_MACHINE_NAME == "x86_64",
    f"kmodify is not implemented for {NORMALIZED_MACHINE_NAME}",
)


@skip_unless_have_test_kmod
@skip_unless_have_kmodify
class TestCallFunction(LinuxKernelTestCase):
    def assert_called(self, name, args, expected_return_value=None):
        if expected_return_value is None:
            expected_return_value = Object(self.prog, "void")
        before = self.prog[f"drgn_kmodify_test_{name}_called"].read_()
        return_value = call_function(self.prog["drgn_kmodify_test_" + name], *args)
        self.assertEqual(
            self.prog[f"drgn_kmodify_test_{name}_called"].read_(),
            before + 1,
        )
        self.assertIdentical(return_value, expected_return_value)

    def assert_returns(self, name, expected_return_value):
        self.assert_called(name, (), expected_return_value)

    def test_void_return(self):
        self.assert_returns("void_return", Object(self.prog, "void"))

    def test_integer_returns(self):
        for name, return_value in (
            ("signed_char", Object(self.prog, "signed char", -66)),
            ("unsigned_char", Object(self.prog, "unsigned char", 200)),
            ("short", Object(self.prog, "short", -666)),
            ("unsigned_short", Object(self.prog, "unsigned short", 7777)),
            ("int", Object(self.prog, "int", -12345)),
            ("unsigned_int", Object(self.prog, "unsigned int", 54321)),
            ("long", Object(self.prog, "long", -2468013579)),
            ("unsigned_long", Object(self.prog, "unsigned long", 4000000000)),
            ("long_long", Object(self.prog, "long long", -9080706050403020100)),
            (
                "unsigned_long_long",
                Object(self.prog, "unsigned long long", 12345678909876543210),
            ),
        ):
            with self.subTest(name=name):
                self.assert_returns(name + "_return", return_value)

    def test_integer_args(self):
        self.assert_called(
            "signed_args",
            (
                Object(self.prog, "signed char", -66),
                Object(self.prog, "short", -666),
                Object(self.prog, "int", -12345),
                Object(self.prog, "long", -2468013579),
                Object(self.prog, "long long", -9080706050403020100),
            ),
        )
        self.assert_called(
            "unsigned_args",
            (
                Object(self.prog, "unsigned char", 200),
                Object(self.prog, "unsigned short", 7777),
                Object(self.prog, "unsigned int", 54321),
                Object(self.prog, "unsigned long", 4000000000),
                Object(self.prog, "unsigned long long", 12345678909876543210),
            ),
        )

    def test_integer_literal_args(self):
        self.assert_called(
            "signed_args",
            (
                -66,
                -666,
                -12345,
                -2468013579,
                -9080706050403020100,
            ),
        )
        self.assert_called(
            "unsigned_args",
            (
                200,
                7777,
                54321,
                4000000000,
                12345678909876543210,
            ),
        )

    def test_many_args(self):
        self.assert_called(
            "many_args",
            (
                48,
                -66,
                -666,
                -12345,
                -2468013579,
                -9080706050403020100,
                200,
                7777,
                54321,
                4000000000,
                12345678909876543210,
            ),
        )

    def test_enum_returns(self):
        self.assert_returns(
            "enum_return", Object(self.prog, "enum drgn_kmodify_enum", 2)
        )

    def test_enum_args(self):
        args = (
            self.prog["DRGN_KMODIFY_ONE"],
            pass_pointer(Object(self.prog, "enum drgn_kmodify_enum", 2)),
        )
        self.assert_called("enum_args", args)
        self.assertIdentical(
            args[1].object, Object(self.prog, "enum drgn_kmodify_enum", 3)
        )

    def test_pointer_returns(self):
        self.assert_returns(
            "pointer_return", self.prog["drgn_kmodify_test_ptr"].read_()
        )

    def test_pointer_args(self):
        self.assert_called(
            "pointer_args", (self.prog["drgn_kmodify_test_ptr"].read_(),)
        )

    def test_string_args(self):
        for msg, f in (
            ("str", lambda t, s: s),
            ("bytes", lambda t, s: s.encode()),
            (
                "array",
                lambda t, s: Object(
                    self.prog,
                    self.prog.array_type(self.prog.type(t), len(s) + 1),
                    s.encode(),
                ),
            ),
            ("str pointer", lambda t, s: pass_pointer(s)),
            ("bytes pointer", lambda t, s: pass_pointer(s.encode())),
            (
                "array pointer",
                lambda t, s: pass_pointer(
                    Object(
                        self.prog,
                        self.prog.array_type(self.prog.type(t), len(s) + 1),
                        s.encode(),
                    )
                ),
            ),
            (
                "pointer",
                lambda t, s: self.prog[f"drgn_kmodify_test_{t.replace(' ', '_')}_str"],
            ),
        ):
            with self.subTest(msg):
                self.assert_called(
                    "string_args",
                    (
                        f("char", "Hello"),
                        f("signed char", ", "),
                        f("unsigned char", "world"),
                        f("const char", "!"),
                    ),
                )

    def test_integer_out_params(self):
        args = [
            pass_pointer(Object(self.prog, "signed char", -66)),
            pass_pointer(Object(self.prog, "short", -666)),
            pass_pointer(-12345),
            pass_pointer(Object(self.prog, "long", -2468013579)),
            pass_pointer(Object(self.prog, "long long", -9080706050403020100)),
        ]
        self.assert_called("integer_out_params", args)
        self.assertIdentical(
            [ptr.object for ptr in args],
            [
                Object(self.prog, "signed char", 33),
                Object(self.prog, "short", 333),
                Object(self.prog, "int", 23456),
                Object(self.prog, "long", 2222222222),
                Object(self.prog, "long long", 9090909090909090909),
            ],
        )

    def test_array_out_params(self):
        arg = pass_pointer(Object(self.prog, "long [3]", [1, 2, 3]))
        self.assert_called("array_out_params", (arg,))
        self.assertIdentical(arg.object, Object(self.prog, "long [3]", [2, 3, 5]))

    def test_array_out_params_extra(self):
        arg = pass_pointer(Object(self.prog, "long [4]", [1, 2, 3, 100]))
        self.assert_called("array_out_params", (arg,))
        self.assertIdentical(arg.object, Object(self.prog, "long [4]", [2, 3, 5, 100]))

    def test_array_out_params_inferred(self):
        self.assert_called(
            "array_out_params", (Object(self.prog, "long [3]", [1, 2, 3]),)
        )

    def test_many_out_params(self):
        args = [
            pass_pointer(Object(self.prog, "char", 48)),
            pass_pointer(Object(self.prog, "signed char", -66)),
            pass_pointer(Object(self.prog, "short", -666)),
            pass_pointer(Object(self.prog, "int", -12345)),
            pass_pointer(Object(self.prog, "long", -2468013579)),
            pass_pointer(Object(self.prog, "long long", -9080706050403020100)),
            pass_pointer(Object(self.prog, "unsigned char", 200)),
            pass_pointer(Object(self.prog, "unsigned short", 7777)),
            pass_pointer(Object(self.prog, "unsigned int", 54321)),
            pass_pointer(Object(self.prog, "unsigned long", 4000000000)),
            pass_pointer(Object(self.prog, "unsigned long long", 12345678909876543210)),
        ]
        self.assert_called("many_out_params", args)
        self.assertIdentical(
            [arg.object for arg in args],
            [
                Object(self.prog, "char", 16),
                Object(self.prog, "signed char", -22),
                Object(self.prog, "short", -222),
                Object(self.prog, "int", -4115),
                Object(self.prog, "long", -822671193),
                Object(self.prog, "long long", -3026902016801006700),
                Object(self.prog, "unsigned char", 66),
                Object(self.prog, "unsigned short", 2592),
                Object(self.prog, "unsigned int", 18107),
                Object(self.prog, "unsigned long", 1333333333),
                Object(self.prog, "unsigned long long", 4115226303292181070),
            ],
        )


@skip_unless_have_test_kmod
@skip_unless_have_kmodify
class TestCallFunctions(LinuxKernelTestCase):
    def test_multiple_calls(self):
        names = ["void_return", "int_return", "unsigned_long_return"]
        before = {
            name: self.prog[f"drgn_kmodify_test_{name}_called"].read_()
            for name in names
        }
        call_functions(
            self.prog,
            [(f"drgn_kmodify_test_{name}",) for name in names],
        )
        for name in names:
            self.assertEqual(
                self.prog[f"drgn_kmodify_test_{name}_called"].read_(),
                before[name] + 1,
            )

    def test_with_args(self):
        before_signed = self.prog["drgn_kmodify_test_signed_args_called"].read_()
        before_unsigned = self.prog["drgn_kmodify_test_unsigned_args_called"].read_()
        call_functions(
            self.prog,
            [
                (
                    "drgn_kmodify_test_signed_args",
                    Object(self.prog, "signed char", -66),
                    Object(self.prog, "short", -666),
                    Object(self.prog, "int", -12345),
                    Object(self.prog, "long", -2468013579),
                    Object(self.prog, "long long", -9080706050403020100),
                ),
                (
                    "drgn_kmodify_test_unsigned_args",
                    Object(self.prog, "unsigned char", 200),
                    Object(self.prog, "unsigned short", 7777),
                    Object(self.prog, "unsigned int", 54321),
                    Object(self.prog, "unsigned long", 4000000000),
                    Object(self.prog, "unsigned long long", 12345678909876543210),
                ),
            ],
        )
        self.assertEqual(
            self.prog["drgn_kmodify_test_signed_args_called"].read_(),
            before_signed + 1,
        )
        self.assertEqual(
            self.prog["drgn_kmodify_test_unsigned_args_called"].read_(),
            before_unsigned + 1,
        )

    def test_with_out_pointers(self):
        args = [
            pass_pointer(Object(self.prog, "signed char", -66)),
            pass_pointer(Object(self.prog, "short", -666)),
            pass_pointer(-12345),
            pass_pointer(Object(self.prog, "long", -2468013579)),
            pass_pointer(Object(self.prog, "long long", -9080706050403020100)),
        ]
        before = self.prog["drgn_kmodify_test_integer_out_params_called"].read_()
        call_functions(
            self.prog,
            [("drgn_kmodify_test_integer_out_params", *args)],
        )
        self.assertEqual(
            self.prog["drgn_kmodify_test_integer_out_params_called"].read_(),
            before + 1,
        )
        self.assertIdentical(
            [ptr.object for ptr in args],
            [
                Object(self.prog, "signed char", 33),
                Object(self.prog, "short", 333),
                Object(self.prog, "int", 23456),
                Object(self.prog, "long", 2222222222),
                Object(self.prog, "long long", 9090909090909090909),
            ],
        )

    def test_empty(self):
        call_functions(self.prog, [])


@skip_unless_have_test_kmod
@skip_unless_have_kmodify
class TestWriteMemory(LinuxKernelTestCase):
    def test_write_memory(self):
        buf = os.urandom(16)
        write_memory(self.prog, self.prog["drgn_kmodify_test_memory"].address_, buf)
        self.assertEqual(
            self.prog.read(self.prog["drgn_kmodify_test_memory"].address_, len(buf)),
            buf,
        )

    def test_fault(self):
        self.assertRaises(FaultError, write_memory, self.prog, 0, b"asdf")


@skip_unless_have_test_kmod
@skip_unless_have_kmodify
class TestSetClearBit(LinuxKernelTestCase):
    BITMAP_PATH = Path("/sys/module/drgn_test/parameters/kmodify_bitmap")

    def test_set_bit(self):
        self.BITMAP_PATH.write_text("11389675720307120785,15873051463709375381\n")

        set_bit(3, self.prog["drgn_kmodify_test_bitmap"])
        self.assertEqual(
            self.BITMAP_PATH.read_text().strip(),
            "11389675720307120793,15873051463709375381",
        )

        set_bit(125, self.prog["drgn_kmodify_test_bitmap"])
        self.assertEqual(
            self.BITMAP_PATH.read_text().strip(),
            "11389675720307120793,18178894472923069333",
        )

    def test_clear_bit(self):
        self.BITMAP_PATH.write_text("11389675720307120785,15873051463709375381\n")

        clear_bit(66, self.prog["drgn_kmodify_test_bitmap"])
        self.assertEqual(
            self.BITMAP_PATH.read_text().strip(),
            "11389675720307120785,15873051463709375377",
        )

        clear_bit(59, self.prog["drgn_kmodify_test_bitmap"])
        self.assertEqual(
            self.BITMAP_PATH.read_text().strip(),
            "10813214968003697297,15873051463709375377",
        )

    def test_type_error(self):
        self.assertRaises(TypeError, set_bit, 3, self.prog["drgn_kmodify_test_memory"])
        self.assertRaises(
            TypeError,
            clear_bit,
            3,
            self.prog["drgn_kmodify_test_bitmap"].address_,
        )

    def test_fault(self):
        self.assertRaises(FaultError, set_bit, 0, NULL(self.prog, "unsigned long *"))


@skip_unless_have_test_kmod
@skip_unless_have_kmodify
class TestWriteObject(LinuxKernelTestCase):
    def test_python_value(self):
        value = random.randrange(2**31)
        write_object(self.prog["drgn_kmodify_test_int"], value)
        self.assertEqual(self.prog["drgn_kmodify_test_int"].value_(), value)

    def test_object_value(self):
        value = random.randrange(2**31)
        write_object(
            self.prog["drgn_kmodify_test_int"], Object(self.prog, "long", value)
        )
        self.assertEqual(self.prog["drgn_kmodify_test_int"].value_(), value)

    def test_pointer_dereference(self):
        value = random.randrange(2**31)
        write_object(
            self.prog["drgn_kmodify_test_int"].address_of_(), value, dereference=True
        )
        self.assertEqual(self.prog["drgn_kmodify_test_int"].value_(), value)

    def test_pointer_no_dereference(self):
        write_object(
            self.prog["drgn_kmodify_test_int_ptr"],
            self.prog["drgn_kmodify_test_int"].address_of_(),
            dereference=False,
        )
        self.assertEqual(
            self.prog["drgn_kmodify_test_int_ptr"],
            self.prog["drgn_kmodify_test_int"].address_of_(),
        )
        write_object(self.prog["drgn_kmodify_test_int_ptr"], 0, dereference=False)
        self.assertEqual(self.prog["drgn_kmodify_test_int_ptr"].value_(), 0)

    def test_pointer_ambiguous(self):
        self.assertRaisesRegex(
            TypeError,
            "use dereference",
            write_object,
            self.prog["drgn_kmodify_test_int_ptr"],
            0,
        )

    def test_not_pointer(self):
        self.assertRaisesRegex(
            TypeError,
            "not a pointer",
            write_object,
            self.prog["drgn_kmodify_test_int"],
            0,
            dereference=True,
        )

    def test_bit_field_size_1(self):
        for i in range(3):
            with self.subTest(i=i):
                write_object(self.prog["drgn_kmodify_test_bit_field"].bit, i)
                self.assertFalse(self.prog["drgn_kmodify_test_bit_field"].expect0_1)
                self.assertTrue(self.prog["drgn_kmodify_test_bit_field"].expect1_1)
                self.assertEqual(
                    self.prog["drgn_kmodify_test_bit_field"].bit.value_(), i % 2
                )
                self.assertFalse(self.prog["drgn_kmodify_test_bit_field"].expect0_2)
                self.assertTrue(self.prog["drgn_kmodify_test_bit_field"].expect1_2)

    def test_byte_aligned_bit_field(self):
        for i in range(3):
            with self.subTest(i=i):
                write_object(self.prog["drgn_kmodify_test_bit_field"].byte_aligned, i)
                self.assertFalse(self.prog["drgn_kmodify_test_bit_field"].expect0_1)
                self.assertTrue(self.prog["drgn_kmodify_test_bit_field"].expect1_1)
                self.assertEqual(
                    self.prog["drgn_kmodify_test_bit_field"].byte_aligned.value_(), i
                )
                self.assertFalse(self.prog["drgn_kmodify_test_bit_field"].expect0_2)
                self.assertTrue(self.prog["drgn_kmodify_test_bit_field"].expect1_2)

    def test_unsupported_bit_field(self):
        self.assertRaises(
            NotImplementedError,
            write_object,
            self.prog["drgn_kmodify_test_bit_field"].two_bits,
            1,
        )
        self.assertRaises(
            NotImplementedError,
            write_object,
            self.prog["drgn_kmodify_test_bit_field"].unaligned,
            1,
        )
